# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" 
This module provides the interface to the modules creating a 1D mesh used for 1D simulations. 

Centerlines can be computed or read in. If centerline are read from a file then they must have 
been split and grouped along branches. If you are reading the centerlines computed by the SimVascular 
Models plugin then the Full_Centerlines.vtp file must be used.
"""
import argparse
import sys
import os

from .manage import get_logger_name, init_logging, get_log_file_name
from .parameters import Parameters
from .centerlines import *
from .mesh import *
from .utils import write_polydata, read_polydata

logger = logging.getLogger(get_logger_name())


class Args(object):
    """ This class defines the command line arguments to the generate-1d-mesh script.
    """
    PREFIX = "--"
    BOUNDARY_SURFACE_DIR = "boundary_surfaces_directory"
    CENTERLINE_INPUT_FILE = "centerlines_input_file"
    CENTERLINE_OUTPUT_FILE = "centerlines_output_file"
    COMPUTE_CENTERLINES = "compute_centerlines"
    COMPUTE_MESH = "compute_mesh"
    DENSITY = "density"
    ELEMENT_SIZE = "element_size"
    INLET_FACE_INPUT_FILE = "inlet_face_input_file"
    INFLOW_INPUT_FILE = "inflow_input_file"
    LINEAR_MATERIAL_EHR = "linear_material_ehr"
    LINEAR_MATERIAL_PRESSURE = "linear_material_pressure"
    MATERIAL_MODEL = "material_model"
    MESH_OUTPUT_FILE = "mesh_output_file"
    MINIMUM_NUMBER_ELEMENTS = "min_num_elements"
    MODEL_NAME = "model_name"
    MODEL_ORDER = "model_order"
    NUM_TIME_STEPS = "num_time_steps"
    OLUFSEN_MATERIAL_K1 = "olufsen_material_k1"
    OLUFSEN_MATERIAL_K2 = "olufsen_material_k2"
    OLUFSEN_MATERIAL_K3 = "olufsen_material_k3"
    OLUFSEN_MATERIAL_EXP = "olufsen_material_exp"
    OLUFSEN_MATERIAL_PRESSURE = "olufsen_material_pressure"
    OUTFLOW_BC_INPUT_FILE = "outflow_bc_input_file"
    OUTFLOW_BC_TYPE = "outflow_bc_type"
    OUTLET_FACE_NAMES_INPUT_FILE = "outlet_face_names_input_file"
    OUTPUT_DIRECTORY = "output_directory"
    SEGMENT_MINIMUM_NUMBER = "seg_min_num"
    SEGMENT_SIZE = "seg_size"
    SEGMENT_SIZE_ADAPTIVE = "seg_size_adaptive"
    SOLVER_OUTPUT_FILE = "solver_output_file"
    SAVE_DATA_FREQUENCY = "save_data_frequency"
    SURFACE_MODEL = "surface_model"
    TIME_STEP = "time_step"
    UNIFORM_BC = "uniform_bc"
    UNITS = "units"
    VISCOSITY = "viscosity"
    WALL_PROPERTIES_INPUT_FILE = "wall_properties_input_file"
    WALL_PROPERTIES_OUTPUT_FILE = "wall_properties_output_file"


def cmd(name):
    """ Create an argparse command argument.
    """
    return Args.PREFIX + name.replace("_", "-")


def parse_args():
    """ Parse command-line arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument(cmd(Args.MODEL_ORDER),
                        help="Dimension of reduced order model (0 or 1)")

    parser.add_argument(cmd(Args.BOUNDARY_SURFACE_DIR),
                        help="Directory containing the boundary (inlet/outlet) surface files")

    parser.add_argument(cmd(Args.CENTERLINE_INPUT_FILE),
                        help="The name of the file to read centerline geometry from.")

    parser.add_argument(cmd(Args.CENTERLINE_OUTPUT_FILE),
                        help="The name of the file to write centerline geometry to.")

    parser.add_argument(cmd(Args.COMPUTE_CENTERLINES), const=True, nargs='?', default=False,
                        help="If given or value is set to 1 then compute centerlines.")

    parser.add_argument(cmd(Args.COMPUTE_MESH), const=True, nargs='?', default=False,
                        help="If given or value is set to 1 then compute a mesh from centerlines..")

    parser.add_argument(cmd(Args.DENSITY),
                        help="The fluid density.")

    parser.add_argument(cmd(Args.ELEMENT_SIZE),
                        help="The finite element size in a sub-segment.")

    parser.add_argument(cmd(Args.INFLOW_INPUT_FILE),
                        help="The name of the file to read inflow data from.")

    parser.add_argument(cmd(Args.INLET_FACE_INPUT_FILE),
                        help="The name of the file (.vtp) defining the inlet face geometry.")

    parser.add_argument(cmd(Args.LINEAR_MATERIAL_EHR),
                        help="Linear material ehr parameter.")
    parser.add_argument(cmd(Args.LINEAR_MATERIAL_PRESSURE),
                        help="Linear material pressure parameter.")

    parser.add_argument(cmd(Args.MATERIAL_MODEL),
                        help="The name of the material model. Can be 'LINEAR' or 'OLUFSEN'.")

    parser.add_argument(cmd(Args.MESH_OUTPUT_FILE),
                        help="The name of the file to write the mesh to.")

    parser.add_argument(cmd(Args.MINIMUM_NUMBER_ELEMENTS),
                        help="The minimum number of finite elements per segment.")

    parser.add_argument(cmd(Args.MODEL_NAME), required=True,
                        help="The name of the model.")

    parser.add_argument(cmd(Args.MODEL_ORDER), required=True,
      help="The order of the model: 0 or 1.")

    parser.add_argument(cmd(Args.NUM_TIME_STEPS),
                        help="The number of simulation time steps.")

    parser.add_argument(cmd(Args.OLUFSEN_MATERIAL_K1),
                        help="Olufsen material k1 parameter.")
    parser.add_argument(cmd(Args.OLUFSEN_MATERIAL_K2),
                        help="Olufsen material k2 parameter.")
    parser.add_argument(cmd(Args.OLUFSEN_MATERIAL_K3),
                        help="Olufsen material k3 parameter.")
    parser.add_argument(cmd(Args.OLUFSEN_MATERIAL_EXP),
                        help="Olufsen material exponent parameter.")
    parser.add_argument(cmd(Args.OLUFSEN_MATERIAL_PRESSURE),
                        help="Olufsen material pressure parameter.")

    parser.add_argument(cmd(Args.OUTFLOW_BC_INPUT_FILE),
                        help="The output boundary condition input file (RESISTANCE, RCR).")

    parser.add_argument(cmd(Args.OUTFLOW_BC_TYPE),
                        help="The output boundary condition type (RESISTANCE, RCR).")

    parser.add_argument(cmd(Args.OUTPUT_DIRECTORY), required=True,
                        help="The directory where output files are written.")

    parser.add_argument(cmd(Args.SEGMENT_MINIMUM_NUMBER),
                        help="The minimum number of sub-segments per vessel master-segment.")

    parser.add_argument(cmd(Args.SEGMENT_SIZE),
                        help="The sub-segment size in a vessel master-segment.")

    parser.add_argument(cmd(Args.SEGMENT_SIZE_ADAPTIVE), const=True, nargs='?', default=False,
                        help="Should the sub-segment size be adaptive to centerline area change?")

    parser.add_argument(cmd(Args.SAVE_DATA_FREQUENCY),
                        help="The frequency to save data as the number of time steps between saves.")

    parser.add_argument(cmd(Args.SOLVER_OUTPUT_FILE),
                        help="The name of the file to write the solver input to.")

    parser.add_argument(cmd(Args.SURFACE_MODEL),
                        help="The surface model used to compute centerlines.")

    parser.add_argument(cmd(Args.TIME_STEP),
                        help="The simulation time step.")

    parser.add_argument(cmd(Args.UNIFORM_BC),
                        help="If set to (true,1,on) then read BC files")

    parser.add_argument(cmd(Args.UNITS),
                        help="The units used to scale geometry. (cm or mm)")

    parser.add_argument(cmd(Args.VISCOSITY),
                        help="The fluid viscosity.")

    parser.add_argument(cmd(Args.OUTLET_FACE_NAMES_INPUT_FILE),
                        help="The file containing outlet face names.")

    parser.add_argument(cmd(Args.WALL_PROPERTIES_INPUT_FILE),
                        help="The name of the file read surface wall material properties from.")

    parser.add_argument(cmd(Args.WALL_PROPERTIES_OUTPUT_FILE),
                        help="The name of the file write grouped wall material properties to.")

    return parser.parse_args(), parser.print_help


def set_parameters(**kwargs):
    """ Set the values of parameters input from the command line.

    The **kwargs argument can originate from the command line or from a direct
    call to the 'run()' function, in which case parser.parse_args() is not called
    and so we need to check here if required arguments have been passed.
    """
    print(kwargs)
    logger.info("Parse arguments ...")

    ## Create a Parameters object to store parameters.
    params = Parameters()

    ## Process arguments.
    #
    true_values = ["on", "true", "1"]

    if kwargs.get(Args.MODEL_ORDER) is not None:
        params.model_order = int(kwargs.get(Args.MODEL_ORDER))
        logger.info("Model order: %s" % params.model_order)
        if params.model_order not in [0, 1]:
            logger.error("The model order '%s' is unknown (must be 0 or 1)." % params.model_order)
            return None
    else:
        logger.error("The model order must be given (0 or 1).")
        return None

    if kwargs.get(Args.BOUNDARY_SURFACE_DIR):
        params.boundary_surfaces_dir = kwargs.get(Args.BOUNDARY_SURFACE_DIR)
        logger.info("Surface directory: %s" % params.boundary_surfaces_dir)
        if not os.path.exists(params.boundary_surfaces_dir):
            logger.error("The surface directory '%s' was not found." % params.boundary_surfaces_dir)
            return None

    if kwargs.get(Args.CENTERLINE_INPUT_FILE):
        params.centerlines_input_file = kwargs.get(Args.CENTERLINE_INPUT_FILE)
        logger.info("Centerlines input file: %s" % params.centerlines_input_file)
        if not os.path.exists(params.centerlines_input_file):
            logger.error("The centerlines input file '%s' was not found." % params.centerlines_input_file)
            return None

    if kwargs.get(Args.CENTERLINE_OUTPUT_FILE):
        params.centerlines_output_file = kwargs.get(Args.CENTERLINE_OUTPUT_FILE)
        logger.info("Centerlines output file: %s" % params.centerlines_output_file)

    # The 'compute_centerlines' parameter is set to True if the COMPUTE_CENTERLINES 
    # argument is given with no value. Otherwise it is set to the value given.
    params.compute_centerlines = (kwargs.get(Args.COMPUTE_CENTERLINES) == True) or \
                                 (kwargs.get(Args.COMPUTE_CENTERLINES) in true_values)
    logger.info("Compute centerlines: %s" % params.compute_centerlines)

    # The 'compute_mesh' parameter is set to True if the COMPUTE_MESH 
    # argument is given with no value. Otherwise it is set to the value given.
    params.compute_mesh = (kwargs.get(Args.COMPUTE_MESH) == True) or \
                          (kwargs.get(Args.COMPUTE_MESH) in true_values)

    if kwargs.get(Args.DENSITY):
        params.density = float(kwargs.get(Args.DENSITY))
    logger.info("Density: %f" % params.density)

    if kwargs.get(Args.ELEMENT_SIZE):
        params.element_size = float(kwargs.get(Args.ELEMENT_SIZE))
    logger.info("Element size: %f" % params.element_size)

    if kwargs.get(Args.INFLOW_INPUT_FILE):
        params.inflow_input_file = kwargs.get(Args.INFLOW_INPUT_FILE)
        logger.info("Inflow input file: %s" % params.inflow_input_file)
        if not os.path.exists(params.inflow_input_file):
            logger.error("The inflow input file '%s' was not found." % params.inflow_input_file)
            return None

    if kwargs.get(Args.INLET_FACE_INPUT_FILE):
        params.inlet_face_input_file = kwargs.get(Args.INLET_FACE_INPUT_FILE)
        if os.path.dirname(params.inlet_face_input_file):
            logger.error("The inlet face input file '%s' should not have a full path." % params.inlet_face_input_file)
            return None
        inlet_file = os.path.join(params.boundary_surfaces_dir, params.inlet_face_input_file)
        if not os.path.exists(inlet_file):
            logger.error("The inlet face input file '%s' was not found in the boundary surfaces diretory '%s'." % \
                         (params.inlet_face_input_file, params.boundary_surfaces_dir))
            return None
        logger.info("Inlet face input file: %s" % params.inlet_face_input_file)

    if kwargs.get(Args.LINEAR_MATERIAL_EHR):
        params.linear_material_ehr = float(kwargs.get(Args.LINEAR_MATERIAL_EHR))
        logger.info("Linear material parameter Er/h: %f" % params.linear_material_ehr)

    if kwargs.get(Args.LINEAR_MATERIAL_PRESSURE):
        params.linear_material_pressure = float(kwargs.get(Args.LINEAR_MATERIAL_PRESSURE))
        logger.info("Linear material parameter pressure: %f" % params.linear_material_pressure)

    if kwargs.get(Args.OUTPUT_DIRECTORY):
        params.output_directory = kwargs.get(Args.OUTPUT_DIRECTORY)
        if not os.path.exists(params.output_directory):
            logger.error("The output directory '%s' was not found." % params.output_directory)
            return None
    else:
        logger.error("An output directory argument must be given.")
        return None

    if kwargs.get(Args.MATERIAL_MODEL):
        params.material_model = kwargs.get(Args.MATERIAL_MODEL)
    logger.info("Material model: %s" % params.material_model)

    if kwargs.get(Args.MESH_OUTPUT_FILE):
        params.mesh_output_file = kwargs.get(Args.MESH_OUTPUT_FILE)
        logger.info("Mesh output file: %s" % params.mesh_output_file)

    if kwargs.get(Args.MINIMUM_NUMBER_ELEMENTS):
        params.min_num_elems = kwargs.get(Args.MINIMUM_NUMBER_ELEMENTS)
    logger.info("Minimum number of finite elements per segment: %d" % params.min_num_elems)

    params.model_name = kwargs.get(Args.MODEL_NAME)
    logger.info("Model name: %s" % params.model_name)

    params.model_order = int(kwargs.get(Args.MODEL_ORDER))
    logger.info("Model order: %s" % params.model_order)

    if kwargs.get(Args.NUM_TIME_STEPS):
        params.num_time_steps = int(kwargs.get(Args.NUM_TIME_STEPS))
    logger.info("Number of time steps: %d" % params.num_time_steps)

    if kwargs.get(Args.OLUFSEN_MATERIAL_K1):
        params.olufsen_material_k1 = float(kwargs.get(Args.OLUFSEN_MATERIAL_K1))
        logger.info("Olufsen material parameter k1: %f" % params.olufsen_material_k1)

    if kwargs.get(Args.OLUFSEN_MATERIAL_K2):
        params.olufsen_material_k2 = float(kwargs.get(Args.OLUFSEN_MATERIAL_K2))
        logger.info("Olufsen material parameter k2: %f" % params.olufsen_material_k2)

    if kwargs.get(Args.OLUFSEN_MATERIAL_K3):
        params.olufsen_material_k3 = float(kwargs.get(Args.OLUFSEN_MATERIAL_K3))
        logger.info("Olufsen material parameter k3: %f" % params.olufsen_material_k3)

    if kwargs.get(Args.OLUFSEN_MATERIAL_EXP):
        params.olufsen_material_exponent = float(kwargs.get(Args.OLUFSEN_MATERIAL_EXP))
        logger.info("Olufsen material parameter exponent: %f" % params.olufsen_material_exponent)

    if kwargs.get(Args.OLUFSEN_MATERIAL_PRESSURE):
        params.olufsen_material_pressure = float(kwargs.get(Args.OLUFSEN_MATERIAL_PRESSURE))
        logger.info("Olufsen material parameter pressure: %f" % params.olufsen_material_pressure)

    if kwargs.get(Args.OUTFLOW_BC_INPUT_FILE):
        params.outflow_bc_file = kwargs.get(Args.OUTFLOW_BC_INPUT_FILE)
        if not os.path.exists(params.outflow_bc_file):
            logger.error("The outflow input bc folder '%s' was not found." % params.outflow_bc_file)
            return None
        logger.info("Outflow bc file: %s" % params.outflow_bc_file)

    if kwargs.get(Args.OUTFLOW_BC_TYPE):
        params.outflow_bc_type = kwargs.get(Args.OUTFLOW_BC_TYPE).split(",")
        logger.info("Outflow bc type: %s" % str(params.outflow_bc_type))

    if kwargs.get(Args.OUTLET_FACE_NAMES_INPUT_FILE):
        params.outlet_face_names_file = kwargs.get(Args.OUTLET_FACE_NAMES_INPUT_FILE)
        if not os.path.exists(params.outlet_face_names_file):
            logger.error("The input outlet face names file '%s' was not found." % params.outlet_face_names_file)
            return None
        logger.info("Outlet face names file: '%s'." % params.outlet_face_names_file)

    if kwargs.get(Args.SEGMENT_MINIMUM_NUMBER):
        params.seg_min_num = int(kwargs.get(Args.SEGMENT_MINIMUM_NUMBER))
    logger.info("Minimum number of sub-segments per master-segment: %d" % params.seg_min_num)

    if kwargs.get(Args.SEGMENT_SIZE):
        params.seg_size = kwargs.get(Args.SEGMENT_SIZE)
    logger.info("Sub-segment size in vessel master-segment: %d" % params.seg_size)

    if kwargs.get(Args.SEGMENT_SIZE_ADAPTIVE):
        params.seg_size_adaptive = int(kwargs.get(Args.SEGMENT_SIZE_ADAPTIVE))
    logger.info("Sub-segment size is adaptive: %d" % params.seg_size_adaptive)

    if kwargs.get(Args.SAVE_DATA_FREQUENCY):
        params.save_data_freq = int(kwargs.get(Args.SAVE_DATA_FREQUENCY))
    logger.info("Save data frequency: %d" % params.save_data_freq)

    if kwargs.get(Args.SOLVER_OUTPUT_FILE):
        params.solver_output_file = kwargs.get(Args.SOLVER_OUTPUT_FILE)
        logger.info("Solver output file: %s" % params.solver_output_file)

    if kwargs.get(Args.SURFACE_MODEL):
        params.surface_model = kwargs.get(Args.SURFACE_MODEL)
        logger.info("Surface model: %s" % params.surface_model)
        if not os.path.exists(params.surface_model):
            logger.error("The surface model file '%s' was not found." % params.surface_model)
            return None

    if kwargs.get(Args.TIME_STEP):
        params.time_step = float(kwargs.get(Args.TIME_STEP))
    logger.info("Simulation time step: %f" % params.time_step)

    if kwargs.get(Args.UNIFORM_BC):
        params.uniform_bc = (kwargs.get(Args.UNIFORM_BC) in true_values)
    logger.info("Uniform boundary conditions: %s" % params.uniform_bc)

    if kwargs.get(Args.UNITS):
        units = kwargs.get(Args.UNITS)
        if params.set_units(units):
            params.units = units
        else:
            logger.error("The units value '%s' was not recognized. Valid values are mm or cm." % units)
            return None
    logger.info("Units: %s" % params.units)

    if kwargs.get(Args.VISCOSITY):
        params.viscosity = float(kwargs.get(Args.VISCOSITY))
    logger.info("Viscosity: %f" % params.viscosity)

    if kwargs.get(Args.WALL_PROPERTIES_INPUT_FILE):
        params.wall_properties_input_file = kwargs.get(Args.WALL_PROPERTIES_INPUT_FILE)
        logger.info("Wall properties input file: %s" % params.wall_properties_input_file)
        if not os.path.exists(params.wall_properties_input_file):
            logger.error("The wall properties input file '%s' was not found." %
                         params.wall_properties_input_file)
            return None
        params.uniform_material = False
        logger.info("Wall properties are not uniform.")
    else:
        logger.info("Wall properties are uniform.")

    if kwargs.get(Args.WALL_PROPERTIES_OUTPUT_FILE):
        params.wall_properties_output_file = kwargs.get(Args.WALL_PROPERTIES_OUTPUT_FILE)
        logger.info("Wall properties output file: %s" % params.wall_properties_output_file)

    ## Check for argument consistency.
    #
    if params.compute_centerlines and params.centerlines_input_file:
        logger.error("Both compute centerlines and read centerlines are given.")
        return None

    if params.compute_centerlines and not params.inlet_face_input_file:
        logger.error("An inlet face file must be given when computing centerlines.")
        return None

    if params.wall_properties_input_file and not params.wall_properties_output_file:
        logger.error("If a wall properties input file is given then a wall properties output file must also be given.")
        return None

    if params.outflow_bc_type and not params.outflow_bc_file:
        logger.error(
            "If an outflow boundary condition type is given then an input data file for that type must also be given.")
        return None

    if not params.outlet_face_names_file:
        logger.error("No outlet face names file was given.")
        return None

    # Uniform / Non-uniform BCs.
    if params.uniform_bc:
        if params.outflow_bc_type != None:
            logger.error("An outflow BC type can't be given with uniform BCs.")
            return None
    else:
        if params.outflow_bc_type == None:
            logger.error("An outflow BC type must be given for non-uniform BCs.")
            return None

    return params


def read_centerlines(params):
    """ Read centerlines for a surface model from a file.
    
    The centerlines must have had 
    """
    centerlines = Centerlines()
    centerlines.read(params, params.centerlines_input_file)

    logger.info("Read centerlines from the file: %s", params.centerlines_input_file)
    logger.info("   Number of points: %d ", centerlines.geometry.GetNumberOfPoints())
    logger.info("   Number of cells: %d ", centerlines.geometry.GetNumberOfCells())
    logger.info("   Number of arrays: %d", centerlines.geometry.GetCellData().GetNumberOfArrays())

    return centerlines


def compute_centerlines(params):
    """ Compute the centerlines for a surface model.
    """

    ## Check input parameters.
    #
    if not params.surface_model:
        logger.error("No surface model has been specified.")
        return

    if not params.centerlines_output_file:
        logger.error("No centerlines output file has been specified.")
        return

    # Create Centerlines object that encapsulats centerline calculations. 
    centerlines = Centerlines()

    # Extract centerlines.
    centerlines.extract_center_lines(params)

    # Write the centerlines branches to a file.
    if centerlines.geometry:
        write_polydata(params.centerlines_output_file, centerlines.geometry)

    return centerlines


def run(**kwargs):
    """ Execute the 1D mesh generation using passed parameters.
    """
    result = ""
    ## Set input parameters.
    params = set_parameters(**kwargs)
    if not params:
        logger.error("Error in parameters.")
        return result
    np.save(os.path.join(params.output_directory, 'parameters'), params)

    centerlines = None

    ## Extract surface centerlines.
    if params.compute_centerlines:
        centerlines = compute_centerlines(params)

    ## Read surface centerlines.
    elif params.centerlines_input_file:
        centerlines = read_centerlines(params)

    if not centerlines:
        logger.error("No centerlines calculated or read in.")
        return result

    ## Generate a 1D mesh.
    mesh = Mesh()
    mesh.generate(params, centerlines)

    # statistics
    num_seg = 0
    num_node = 0
    for i in mesh.cell_data['id'].values():
        num_seg += i.shape[0]
        num_node += num_seg + 1

    logger.info("Generated %d segments, %d nodes and %d elements." % (num_seg, num_node, mesh.num_elements))
    result = "Mesh: num_nodes=%d num_elements=%d num_segs=%d\n" % (num_node, mesh.num_elements, num_seg)
    return result


def run_from_c(*args, **kwargs):
    """ Execute the 1D mesh generation using passed parameters from c++

    The '*args' argument contains the directory to write the log file.
    """
    output_dir = args[0]
    init_logging(output_dir)
    msg = "Status: OK\n"

    try:
        result = run(**kwargs)
        msg += result
    except Exception as e:
        logger.error(str(e))
        msg = "Status: Error\n"
        msg += "Error: " + str(e) + "\n"

    ## Attach log file to returned result.
    #
    msg += "Log:\n"
    log_file_name = os.path.join(output_dir, get_log_file_name())

    with open(log_file_name, 'r') as file:
        msg += file.read()

    return msg


if __name__ == '__main__':
    init_logging()
    args, print_help = parse_args()
    if not run(**vars(args)):
        sys.exit(1)



import threading
from threading import Lock, Condition
from scipy.spatial import cKDTree
import numpy as np
from time import perf_counter
from threading import Lock, Condition
import hnswlib
import os
from usearch.index import Index, MetricKind
MAX_THREADS = min(4, max(1, os.cpu_count() - 1))
# TODO: Consider using the hnswlib library for approximate nearest neighbor for large datasets (n >= 1e5)
# TODO: Consider Usearch library for approximate nearest neighbor

class USearchTree:
    def __init__(self, data, space='l2sq', ef_construction=200, M=100, ef=20):
        """
        Initialize USearch tree with the given data.

        Parameters:
        - data: A 2D numpy array of shape (N, D), where N is the number of points, and D is the dimensionality.
        - space: Metric to use for distance calculation. Options: 'l2' (Euclidean), 'ip' (inner product), 'cos' (cosine similarity).
        - ef_construction: The size of the dynamic candidate list during construction.
        - M: Controls the number of bi-directional links created for every new element during construction.
        """
        self.dim = data.shape[1]
        metric = {'l2': 'euclidean', 'ip': 'dot', 'cos': 'cos'}.get(space, 'euclidean')

        self.index = Index(
            ndim=self.dim,
            metric=space,
            connectivity=M,
            expansion_add=ef_construction,
            expansion_search=ef,
            dtype='f32'
        )

        labels = np.arange(data.shape[0])
        self.index.add(labels, data)

    def query(self, points, k=1, threads=MAX_THREADS):
        """
        Query the nearest neighbors for the given points.

        Parameters:
        - points: A 2D numpy array of shape (N, D) where N is the number of query points.
        - k: The number of nearest neighbors to return.

        Returns:
        - distances: A 2D array of distances to the nearest neighbors.
        - indices: A 2D array of indices of the nearest neighbors.
        """
        results = self.index.search(points, k, threads=threads)
        distances = results.distances
        indices = results.keys
        return distances, indices

    def query_ball_point(self, points, r, **kwargs):
        """
        Query points within distance `r` from the given points.

        Parameters:
        - points: A 2D numpy array of shape (N, D) where N is the number of query points.
        - r: The radius to search within.

        Returns:
        - A list of lists, where each sublist contains the indices of points within distance `r` of the corresponding query point.
        """
        # Approximate a ball query by performing knn search with a large `k` and filtering by `r`
        k = kwargs.get("k", min(50, self.index.size))
        distances, indices = self.query(points, k)

        # Filter based on radius `r`
        result = []
        for i in range(len(points)):
            within_r = indices[i][distances[i] <= r]  # Filter neighbors by radius
            result.append(within_r)

        return result

    def set_ef(self, ef):
        """
        Set the `ef` parameter, which controls the search depth and the trade-off between speed and accuracy.

        Parameters:
        - ef: The size of the dynamic list for nearest neighbors search.
        """
        self.index.expansion_search = ef

    def add_items(self, data, ids):
        """
        Add more points to the USearch index.

        Parameters:
        - data: A 2D numpy array of new points to add.
        """
        self.index.add(ids, data)

    def replace(self, data, ids):
        """
        Replace points in the index.

        Parameters:
        - data: A 2D numpy array of new points to replace existing points.
        - ids: A list of IDs of the points to replace.
        """
        self.index.remove(ids)
        self.index.add(ids, data)

class HNSWTree:
    def __init__(self, data, space='l2', ef_construction=200, M=16, num_elements=int(4e6), ef=20):
        """
        Initialize HNSW tree with the given data.

        Parameters:
        - data: A 2D numpy array of shape (N, D), where N is the number of points, and D is the dimensionality.
        - space: Metric to use for distance calculation. Options: 'l2' (Euclidean), 'ip' (inner product).
        - ef_construction: The size of the dynamic candidate list during construction.
        - M: Controls the number of bi-directional links created for every new element during construction.
        """
        self.num_elements = num_elements
        _, self.dim = data.shape
        self.index = hnswlib.Index(space=space, dim=self.dim)
        self.index.init_index(max_elements=self.num_elements, ef_construction=ef_construction, M=M,
                              allow_replace_deleted=True)
        threads = os.cpu_count() // 2 - 1
        self.index.set_num_threads(threads)
        labels = np.arange(data.shape[0])
        self.index.add_items(data, labels)

        # Control the trade-off between accuracy and speed at query time
        self.index.set_ef(ef)

    def query(self, points, k=1):
        """
        Query the nearest neighbors for the given points.

        Parameters:
        - points: A 2D numpy array of shape (N, D) where N is the number of query points.
        - k: The number of nearest neighbors to return.

        Returns:
        - distances: A 2D array of distances to the nearest neighbors.
        - indices: A 2D array of indices of the nearest neighbors.
        """
        indices, distances = self.index.knn_query(points, k=k)
        return distances, indices

    def query_ball_point(self, points, r, **kwargs):
        """
        Query points within distance `r` from the given points.

        Parameters:
        - points: A 2D numpy array of shape (N, D) where N is the number of query points.
        - r: The radius to search within.

        Returns:
        - A list of lists, where each sublist contains the indices of points within distance `r` of the corresponding query point.
        """
        # Approximate a ball query by performing knn search with a large `k` and filtering by `r`
        k = kwargs.get("k", min(50, self.index.element_count))  # A large `k` value to ensure enough neighbors are found
        indices, distances = self.index.knn_query(points, k=k)

        # Filter based on radius `r`
        result = []
        for i in range(len(points)):
            within_r = indices[i][distances[i] <= r].tolist()  # Filter neighbors by radius
            result.append(within_r)

        return result

    def set_ef(self, ef):
        """
        Set the `ef` parameter, which controls the search depth and the trade-off between speed and accuracy.

        Parameters:
        - ef: The size of the dynamic list for nearest neighbors search.
        """
        self.index.set_ef(ef)

    def add_items(self, data, ids):
        """
        Add more points to the HNSW index.

        Parameters:
        - data: A 2D numpy array of new points to add.
        """
        self.index.add_items(data, ids, num_threads=1)
        self.num_elements += data.shape[0]

    def replace(self, data, ids):
        [self.index.mark_deleted(i) for i in ids]
        self.index.add_items(data, ids, num_threads=1, replace_deleted=True)

class ReadWriteLock:
    def __init__(self):
        self._read_ready = Condition(Lock())
        self._readers = 0

    def acquire_read(self):
        with self._read_ready:
            self._readers += 1

    def release_read(self):
        with self._read_ready:
            self._readers -= 1
            if not self._readers:
                self._read_ready.notifyAll()

    def acquire_write(self):
        self._read_ready.acquire()
        while self._readers > 0:
            self._read_ready.wait()

    def release_write(self):
        self._read_ready.release()


class KDTreeManager:
    def __init__(self, data):
        self.active_tree = cKDTree(data)
        self.previous_tree = None
        self.update_tree = None
        self.tree_lock = threading.Lock()
        self.update_done_event = threading.Event()
        self.update_done_event.set()
        self.update_event = threading.Event()
        self.shutdown_event = threading.Event()

        # This variable will hold the latest update request
        self.update_queue = []
        self.update_queue_lock = threading.Lock()

        # Start a background thread that will handle updates
        self.update_thread = threading.Thread(target=self._update_worker, daemon=True)
        self.update_thread.start()

    def start_update(self, new_data):
        with self.update_queue_lock:
            # Store the new data as the latest pending update
            self.update_queue.append(new_data)

        # If there's no current update thread or it's finished, start a new one
        #if self.update_thread is None or not self.update_thread.is_alive():
        #    self._start_new_update()
        self.update_done_event.clear()
        self.update_event.set()

    def _update_worker(self):
        while not self.shutdown_event.is_set():
            # Wait for an update event
            self.update_event.wait()

            # Process all updates in the queue
            while True:
                with self.update_queue_lock:
                    if not self.update_queue:
                        break
                    new_data = self.update_queue.pop(-1)
                    self.update_queue = []

                # Build the new KDTree
                new_tree = cKDTree(new_data)
                # Swap the trees safely
                with self.tree_lock:
                    self.update_tree = new_tree

                # Immediately swap active trees if no further updates are queued
                #if not self.update_queue:
                #    self.swap_trees()
                self.update_done_event.set()
            # Clear the event after processing all updates
            self.update_event.clear()

    #def _start_new_update(self):
    #    with self.update_queue_lock:
    #        if self.pending_update is None:
    #            return  # No update to start
    #
    #        # Retrieve and clear the pending update
    #        new_data = self.pending_update
    #        self.pending_update = None
    #
    #    # Start a new update thread
    #    self.update_done_event.clear()
    #    self.update_thread = threading.Thread(target=self._update_tree, args=(new_data,))
    #    self.update_thread.start()

    #def _update_tree(self, new_data):
    #    # Build the new KDTree
    #    new_tree = cKDTree(new_data)
    #    # Swap the trees safely
    #    with self.tree_lock:
    #        self.update_tree = new_tree
    #
    #    # Signal that the update is done
    #    self.update_done_event.set()
    #
    #    # Check if there is another update waiting to be processed
    #    self._start_new_update()

    def swap_trees(self):
        with self.tree_lock:
            if self.update_tree is None:
                return
            self.previous_tree = self.active_tree
            self.active_tree, self.update_tree = self.update_tree, self.active_tree
            self.update_tree = None

    def query(self, points, k=1):
        #with self.tree_lock:
        result = self.active_tree.query(points, k=k)
        return result

    def query_ball_point(self, points, r):
        #with self.tree_lock:
        result = self.active_tree.query_ball_point(points, r)
        return result

    def query_ball_tree(self, other, r, eps=0.0):
        with self.tree_lock:
            result = self.active_tree.query_ball_tree(other, r, eps=eps)
            return result

    def revert_to_previous(self):
        with self.tree_lock:
            if self.previous_tree is not None:
                self.active_tree, self.previous_tree = self.previous_tree, self.active_tree
                print("Reverted to the previous KDTree.")
            else:
                print("No previous KDTree to revert to.")

    def wait_for_update(self):
        while True:
            # Wait for the current update to complete
            self.update_done_event.wait()

            # Check if there's another pending update
            with self.update_queue_lock:
                if not self.update_queue:
                    break  # No more updates, exit loop
                else:
                    self.update_done_event.clear()
                    self.update_event.set()
                    self._update_worker()

    def shutdown(self):
        # Signal the update thread to shut down
        self.shutdown_event.set()
        self.update_event.set()  # Wake up the thread to exit

        # Wait for the thread to finish
        self.update_thread.join()



import numpy as np
#import numba as nb
from scipy.spatial import cKDTree
from scipy import spatial
from tqdm import tqdm
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
from matplotlib import cm
import open3d as o3d
import sklearn.neighbors
from sklearn.preprocessing import PolynomialFeatures
from sklearn.mixture import GaussianMixture


def bases(points):
    n = points.shape[0]
    d = points.shape[1]
    k = int(d*np.round(np.log(n)))
    kdt = spatial.KDTree(points)
    dist, idx = kdt.query(points, k=k+1)
    for i in range(n):
        local_points = points[idx[i,:],:]
        local_points = local_points - np.ones((k+1,1))@local_points[0,:].reshape(1,-1)
        local_points = local_points[1:,:]
        local_points = local_points/np.linalg.norm(local_points,axis=1).reshape(-1,1)
        u, s, vh = np.linalg.svd(local_points)


def generate_bases(points, quiet=True):
    n_points = points.shape[0]
    dim = points.shape[1]
    int_dim = dim - 1
    tangent_basis = np.zeros((dim, int_dim, n_points))
    normal_basis = np.zeros((dim, dim - int_dim, n_points))
    K = int(min(20, dim * np.log10(n_points))) # can have problems with low numbers of points
    u, u_index, u_counts = np.unique(points,axis=1,return_counts=True,return_index=True)
    KDT = spatial.KDTree(u)
    dist, idx = KDT.query(u, k=K + 1)
    count = 0
    if quiet:
        for i in range(u.shape[0]):
            tmp = points[idx[i, :], :]
            # Account for duplicate points for multiple normal estimation
            # and feature edge detection
            # if u_counts[i] > 1:
            # ensure that the count of the points is greater than the
            # number of neighbors to avoid singular matrix solutions
            tmp = tmp - np.ones((K + 1, 1)) @ tmp[0, :].reshape(1, -1)
            tmp = tmp[1:, :]
            U, S, H = np.linalg.svd(tmp)
            H = H.T
            diag_idx = S.argsort()[::-1]
            sort_diag = np.sort(S)[::-1]
            H = H[:, diag_idx]
            tangent_basis[:, :, i] = H[:, :int_dim]
            normal_basis[:, :, i] = H[:, dim - 1].reshape(-1, 1)
    else:
        for i in tqdm(range(n_points), desc='Generating bases       '):
            tmp = points[idx[i, :], :]
            tmp = tmp - np.ones((K + 1, 1)) @ tmp[0, :].reshape(1, -1)
            tmp = tmp[1:, :]
            U, S, H = np.linalg.svd(tmp)
            H = H.T
            diag_idx = S.argsort()[::-1]
            sort_diag = np.sort(S)[::-1]
            H = H[:, diag_idx]
            tangent_basis[:, :, i] = H[:, :int_dim]
            normal_basis[:, :, i] = H[:, dim - 1].reshape(-1, 1)
    return tangent_basis, normal_basis, idx, KDT


def estimate_weingarten_map(points, quiet=True):
    n_points = points.shape[0]
    dim = points.shape[1]
    tangent_basis, normal_basis, idx, KDT = generate_bases(points, quiet=quiet)
    tangent_dim = tangent_basis.shape[1]
    normal_dim = normal_basis.shape[1]
    weingarten_map = np.zeros((tangent_dim, tangent_dim,
                               normal_dim, n_points))
    if quiet:
        for i in range(n_points):
            for j in range(normal_dim):
                tmp_point = points[i, :]
                tmp_tangent_basis = tangent_basis[:, :, i]
                tmp_normal_basis = normal_basis[:, j, i]
                tmp_neighborhood = idx[i, :]
                tmp_neighborhood_size = len(tmp_neighborhood)
                tmp_local_normals = normal_basis[:, :, tmp_neighborhood]
                tmp_normal_extension = np.zeros((tmp_neighborhood_size, dim))
                for k in range(tmp_neighborhood_size):
                    projection = tmp_local_normals[:, :, k] @ \
                                 tmp_local_normals[:, :, k].T @ \
                                 tmp_normal_basis.reshape(-1, 1)
                    tmp_normal_extension[k, :] = projection.T
                tmp_diff_normal = np.zeros((tmp_neighborhood_size - 1, dim))
                tmp_diff_position = np.zeros((tmp_neighborhood_size - 1, dim))
                for k in range(tmp_neighborhood_size - 1):
                    tmp_diff_normal[k, :] = tmp_normal_extension[k + 1, :] - tmp_normal_basis.reshape(1, -1)
                    tmp_diff_position[k, :] = points[tmp_neighborhood[k + 1], :] - tmp_point
                tmp_normal_projection = tmp_diff_normal @ tmp_tangent_basis
                tmp_position_projection = tmp_diff_position @ tmp_tangent_basis
                A = -(np.linalg.inv(tmp_position_projection.T @ tmp_position_projection)) @ \
                    tmp_position_projection.T @ tmp_normal_projection
                weingarten_map[:, :, j, i] = (1 / 2) * (A + A.T)
    else:
        for i in tqdm(range(n_points), desc='Building Weingarten Map'):
            for j in range(normal_dim):
                tmp_point = points[i, :]
                tmp_tangent_basis = tangent_basis[:, :, i]
                tmp_normal_basis = normal_basis[:, j, i]
                tmp_neighborhood = idx[i, :]
                tmp_neighborhood_size = len(tmp_neighborhood)
                tmp_local_normals = normal_basis[:, :, tmp_neighborhood]
                tmp_normal_extension = np.zeros((tmp_neighborhood_size, dim))
                for k in range(tmp_neighborhood_size):
                    projection = tmp_local_normals[:, :, k] @ \
                                 tmp_local_normals[:, :, k].T @ \
                                 tmp_normal_basis.reshape(-1, 1)
                    tmp_normal_extension[k, :] = projection.T
                tmp_diff_normal = np.zeros((tmp_neighborhood_size - 1, dim))
                tmp_diff_position = np.zeros((tmp_neighborhood_size - 1, dim))
                for k in range(tmp_neighborhood_size - 1):
                    tmp_diff_normal[k, :] = tmp_normal_extension[k + 1, :] - \
                                            tmp_normal_basis.reshape(1, -1)
                    tmp_diff_position[k, :] = points[tmp_neighborhood[k + 1], :] - \
                                              tmp_point
                tmp_normal_projection = tmp_diff_normal @ tmp_tangent_basis
                tmp_position_projection = tmp_diff_position @ tmp_tangent_basis
                A = -(np.linalg.inv(tmp_position_projection.T @ \
                                    tmp_position_projection)) @ \
                    tmp_position_projection.T @ \
                    tmp_normal_projection
                weingarten_map[:, :, j, i] = (1 / 2) * (A + A.T)
    return weingarten_map, idx, KDT


def estimate_gaussian_curvature(points, quiet=True):
    n_points = points.shape[0]
    guassian_curvature = np.zeros((n_points, 1))
    w_map, idx, KDT = estimate_weingarten_map(points, quiet=quiet)
    for i in range(n_points):
        guassian_curvature[i] = np.linalg.det(w_map[:, :, 0, i])
    return guassian_curvature, idx, KDT


def show_gaussian_curvature(points):
    guassian_curvature, idx, KDT = estimate_gaussian_curvature(points)
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    coolwarm = cm.get_cmap('coolwarm', 1000)
    gc_min = min(guassian_curvature)
    gc_max = max(guassian_curvature)
    colors = coolwarm(guassian_curvature)
    cax = ax.scatter3D(points[:, 0], points[:, 1], points[:, 2], cmap=coolwarm, c=guassian_curvature)
    cbar = fig.colorbar(cax, ax=ax, ticks=[gc_min, gc_max])
    return fig, ax


def estimate_normals(points,radius=0.1, nn=20):
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=radius, max_nn=nn))
    pcd.orient_normals_consistent_tangent_plane(nn)
    normals = np.asarray(pcd.normals)
    return points, normals


def fit_surface(points,degree=3,nn=20):
    nbrs = sklearn.neighbors.NearestNeighbors(n_neighbors=nn, algorithm='ball_tree').fit(points)
    distances, indices = nbrs.kneighbors(points)

    # fit a polynomial surface to the points

    poly = PolynomialFeatures(degree=degree)
    curvatures = []
    for i in range(points.shape[0]):
        neighbors = points[indices[i,:],:]
        centered_points = neighbors - points[i,:]

        X = poly.fit_transform(centered_points)


import sympy as sp
from sympy.interactive import printing
printing.init_printing(use_latex=True)


class BaseUnit(object):
    def __init__(self, value, base, unit=None):
        """
        The BaseUnit class defines the base units that are
        used to generate a synthetic vascular tree.
        """
        valid_bases = ['length', 'time', 'mass']
        base_dimensions = {'length': ['L'],
                           'time': ['T'],
                           'mass': ['M']}
        default_units = {'length': 'cm',
                         'time': 's',
                         'mass': 'g'}
        if base not in valid_bases:
            raise ValueError("Invalid base units: {}.".format(base))
        else:
            if unit is None:
                unit = default_units[base]
        self.dimensions = [base_dimensions[base], []]
        self.unit = sp.Symbol(unit)
        self.base = base
        self.value = value

    def __str__(self):
        return str(self.value) + ' ' + str(self.unit)

    def __mul__(self, other):
        if isinstance(other, BaseUnit):
            if self.base == other.base:
                conversion = ConvertBaseUnit(self, other)
                new_value = conversion.factor
            return self.unit * other.unit

    def __truediv__(self, other):
        if isinstance(other, BaseUnit):
            return self.unit / other.unit

    def __pow__(self, other):
        if isinstance(other, int):
            return self.unit ** other


class BaseUnitSystem(object):
    def __init__(self, **kwargs):
        """
        The BaseUnits class defines the base units that are
        used to generate a synthetic vascular tree.
        """
        length = kwargs.get('length', 'cm')
        time = kwargs.get('time', 's')
        mass = kwargs.get('mass', 'g')
        self.length = BaseUnit('length', length)
        self.time = BaseUnit('time', time)
        self.mass = BaseUnit('mass', mass)

    def __str__(self):
        output = "Base Units:\n"
        output += "------------\n"
        output += "Length: {}\n".format(self.length.unit)
        output += "  Time: {}\n".format(self.time.unit)
        output += "  Mass: {}\n".format(self.mass.unit)
        output += "------------"
        return output


class DerivedUnit(object):
    def __init__(self, value, unit, dimensions):
        """
        The DerivedUnit class defines the derived units that are
        used to generate a synthetic vascular tree.
        """
        self.value = value
        self.unit = unit
        self.dimensions = dimensions

    def __str__(self):
        return str(self.value) + ' ' + str(self.unit)

    def __mul__(self, other):
        if isinstance(other, DerivedUnit):
            return self.unit * other.unit

class NamedUnit(object):
    def __init__(self, derived, base_unit_system, **kwargs):
        """
        The DerivedUnit class defines the derived units that are
        used to generate a synthetic vascular tree.
        """
        self.base_unit_system = base_unit_system
        valid_derived_types = {'pressure': [['M'], ['L', 'T', 'T']],
                               'density': [['M'], ['L', 'L', 'L']],
                               'volumetric_flow': [['L', 'L', 'L'], ['T']],
                               'mass_flow': [['M'], ['T']],
                               'area': [['L', 'L']],
                               'volume': [['L', 'L', 'L']],
                               'velocity': [['L'], ['T']],
                               'acceleration': [['L'], ['T', 'T']],
                               'kinematic_viscosity': [['L', 'L'], ['T']],
                               'dynamic_viscosity': [['M'], ['L', 'T']],
                               'shear_rate': [[], ['T']],
                               'shear_stress': [['M'], ['L', 'T', 'T']]}
        if derived not in valid_derived_types.keys():
            raise ValueError("Invalid type: {}.".format(derived))
        else:
            self.base = derived
        self.dimensions = valid_derived_types[derived]
        numerator = 1
        denominator = 1
        map = {'L': base_unit_system.length.unit,
               'T': base_unit_system.time.unit,
               'M': base_unit_system.mass.unit}
        for dim in self.dimensions[0]:
            numerator *= map[dim]
        for dim in self.dimensions[1]:
            denominator *= map[dim]
        self.unit = numerator / denominator

    def __str__(self):
        output = "Derived Unit: {}\n".format(self.base)
        output += "------------\n"
        output += "{}\n".format(self.unit)
        output += "------------"
        return output

    def __mul__(self, other):
        if isinstance(other, DerivedUnit):
            return self.unit * other.unit

    def __truediv__(self, other):
        if isinstance(other, DerivedUnit):
            return self.unit / other.unit


class UnitSystem(object):
    def __init__(self, **kwargs):
        """
        The Units class defines the units that are used
        to generate a synthetic vascular tree.
        """
        # Base Units
        self.length = 'cm'
        self.time = 's'
        self.mass = 'g'
        # Derived Units
        self.pressure = None
        self.flow = None
        self.area = None
        self.volume = None
        self.velocity = None
        self.acceleration = None
        self.density = None
        self.kinematic_viscosity = None
        self.dynamic_viscosity = None
        self.shear_rate = None
        self.shear_stress = None


class ConvertBaseUnit(object):
    def __init__(self, base_unit_0, base_unit_1):
        """
        Convert between two base units.
        """
        self.factor = 1
        metric_length = {'nm': 1e-9,
                         'um': 1e-6,
                         'mm': 1e-3,
                         'cm': 1e-2,
                         'm': 1,
                         'km': 1e3}
        metric_weight = {'ng': 1e-12,
                         'ug': 1e-9,
                         'mg': 1e-6,
                         'g': 1e-3,
                         'kg': 1}
        imperial_length = {'in': 1,
                           'ft': 12,
                           'yd': 36,
                           'mi': 63360}
        imperial_weight = {'oz': 1,
                           'lb': 16,
                           'stone': 224,
                           'ton': 35840}
        metric_to_imperial = {'m_to_in': 39.3701,
                              'kg_to_oz': 35.274}
        imperial_to_metric = {'in_to_m': 0.0254,
                              'oz_to_kg': 0.0283495}
        time = {'s': 1,
                'min': 60,
                'hr': 3600,
                'day': 86400,
                'wk': 604800,
                'mo': 2.628e6,
                'yr': 31536000}
        if base_unit_0.base != base_unit_1.base:
            raise ValueError("Base units must be the same type. {} != {}.".format(base_unit_0.base,base_unit_1.base))
        base_unit_0_type = None
        base_unit_1_type = None
        if base_unit_0.base == 'length':
            if str(base_unit_0.unit) in metric_length.keys():
                base_unit_0_type = 'metric'
                self.factor *= metric_length[str(base_unit_0.unit)]
            elif str(base_unit_0.unit) in imperial_length.keys():
                base_unit_0_type = 'imperial'
                self.factor *= imperial_length[str(base_unit_0.unit)]
            else:
                raise ValueError("Invalid first unit: {}.".format(base_unit_0.unit))
            if str(base_unit_1.unit) in metric_length.keys():
                base_unit_1_type = 'metric'
                if base_unit_0_type == base_unit_1_type:
                    self.factor *= 1/metric_length[str(base_unit_1.unit)]
                else:
                    self.factor *= imperial_to_metric['in_to_m']
                    self.factor *= 1/imperial_length[str(base_unit_1.unit)]
            elif str(base_unit_1.unit) in imperial_length.keys():
                base_unit_1_type = 'imperial'
                if base_unit_0_type == base_unit_1_type:
                    self.factor *= 1/imperial_length[str(base_unit_1.unit)]
                else:
                    self.factor *= metric_to_imperial['m_to_in']
                    self.factor *= 1/imperial_length[str(base_unit_1.unit)]
            else:
                raise ValueError("Invalid second unit: {}.".format(base_unit_1.unit))




import numpy as np
from .kernel.kernel import Kernel
from .solver.solver import Solver
from time import perf_counter_ns


class Patch:
    """
    Patch class decomposes a domain into a set of interpolation sub-problems which
    can be solved independently and then blended together to form the final
    domain interpolation function.
    """
    def __init__(self, lam=0):
        self.points = None
        self.normals = None
        self.kernel = None
        self.solver = None
        self.constants = None
        self.rbf_degree = 3
        self.lam = lam

    def set_data(self, *args):
        """
        Set the data for the domain.
        """
        if len(args) > 1:
            self.points = args[0]
            self.normals = args[1]
            self.kernel = Kernel(self.points, lam=self.lam)
            self.kernel.set_initial_values(self.normals)
        elif len(args) > 0:
            self.points = args[0]
            self.kernel = Kernel(self.points, lam=self.lam)
            self.kernel.set_initial_values()
        else:
            print("Error: No data provided.")
        return None

    def solve(self, method="L-BFGS-B", precision=9):
        """
        Solve the interpolation problem for the patch object

        Parameters:
        -----------
        method : str
            The method to use for the solver. Default is L-BFGS-B.
            (Limited-memory Broyden-Fletcher-Goldfarb-Shanno Bounded)

        precision : int
            The number of decimal places to round the constants to.
            Default is 9.

        Returns:
        --------
        None
        """
        self.solver = Solver(self.kernel)
        self.solver.set_solver(method=method)
        if isinstance(self.normals, type(None)):
            self.solver.solve()
        else:
            self.solver.solve(skip=True)
        self.constants = np.round(self.solver.get_constants(), decimals=precision)
        return None

    def build(self):
        """
        Build the interpolation function for the patch object
        :return:
        """
        a = np.array(self.constants[:self.kernel.n])
        b = np.array(self.constants[self.kernel.n:self.kernel.n * (self.kernel.d + 1)].reshape(self.kernel.n, self.kernel.d))
        c = np.array(self.constants[self.kernel.n * (self.kernel.d + 1):self.kernel.n * (self.kernel.d + 1) + self.kernel.d])
        d = np.array(self.constants[-1])
        a = a.reshape(a.shape + (1,) * self.kernel.d)
        b = b.reshape(tuple([b.shape[0]]) + (1,) * self.kernel.d + tuple([b.shape[1]]))
        c = c.reshape((1,) * self.kernel.d + c.shape)
        pts = np.array(self.points.reshape(tuple([self.points.shape[0]]) + (1,) * self.kernel.d + tuple([self.points.shape[1]])))
        def f(x, a_=a, b_=b, c_=c, d_=d, points=pts, show=False):
            """
            Interpolation function for a patch of the domain.

            Parameters:
            -----------
            x : array_like
                The point(s) at which to evaluate the interpolation function. Given as an
                ndarray of shape (..., d), where d is the dimension of the domain.

            a_ : array_like
                The coefficients for the Duchon interpolation function.

            b_ : array_like
                The coefficients for the gradient of the Duchon interpolation function.

            c_ : array_like
                The coefficients for the Hessian of the Duchon interpolation function.

            d_ : float
                The constant term for the Duchon interpolation function.

            Returns:
            --------
            value : array_like
                The value of the interpolation function at the given point(s). Given as an
                ndarray of shape (..., 1).
            """
            diff = x - points
            diff_2 = np.sum(np.square(diff), axis=-1)
            if show:
                print("Diff: ", diff)
                print("Diff^2: ", diff_2)
            #a_value = np.sum(a_ * np.sum(np.square(diff), axis=-1) ** (self.rbf_degree / 2), axis=0)
            # A Value
            value = np.sum(a_ * np.power(diff_2, (self.rbf_degree / 2)), axis=0)
            if show:
                print("A Value: ", value)
            # B Value
            value += np.sum(self.rbf_degree * np.sum((-b_) * diff, axis=-1) * np.power(diff_2, (
                        self.rbf_degree / 2 - 1)), axis=0)
            if show:
                print("A + B Value: ", value)
            # C Value
            value += np.sum(c_ * x, axis=-1)
            if show:
                print("A + B + C Value: ", value)
            # D Value
            value += d_
            if show:
                print("A + B + C + D Value: ", value)
            return value
        f.points = self.points
        f.pts = pts
        if self.normals is not None:
            f.normals = self.normals
        else:
            f.normals = self.solver.get_normals()
        f.dimensions = self.kernel.d
        f.min = np.min(self.points, axis=0)
        f.max = np.max(self.points, axis=0)
        f.centroid = np.mean(self.points, axis=0)
        f.first = self.points[0, :]
        f.a = a
        f.b = b
        f.c = c
        f.d = d
        return f



import pyvista


def show(forest, plot_domain=False, return_plotter=False, **kwargs):
    """
    This function is used to visualize a synthetic vascular tree.
    """
    colors = kwargs.get('colors', ['red', 'blue', 'green', 'yellow', 'purple',
                                   'orange', 'cyan', 'magenta', 'white', 'black'])
    plotter = pyvista.Plotter(**kwargs)
    count = 0
    for network in forest.networks:
        for tree in network:
            for i in range(tree.data.shape[0]):
                center = (tree.data[i, 0:3] + tree.data[i, 3:6]) / 2
                direction = tree.data.get('w_basis', i)
                radius = tree.data.get('radius', i)
                length = tree.data.get('length', i)
                vessel = pyvista.Cylinder(center=center, direction=direction, radius=radius, height=length)
                plotter.add_mesh(vessel, color=colors[count % len(colors)])
            count += 1
    if plot_domain:
        plotter.add_mesh(forest.domain.boundary, color='grey', opacity=0.25)
    if return_plotter:
        return plotter
    else:
        plotter.show()



import pytest
import numpy as np
import pyvista as pv

from svv.domain.routines.boolean import (
    convert_to_trimesh,
    convert_to_pyvista,
    boolean
)


###############################################################################
# Fixtures for test shapes
###############################################################################

@pytest.fixture
def cube_pv():
    """
    A unit cube from (0,0,0) to (1,1,1).
    Triangulate so that every face is made of two triangles.
    """
    box = pv.Box(bounds=(0, 1, 0, 1, 0, 1))
    return box.triangulate()

@pytest.fixture
def sphere_pv():
    """
    A sphere centered at (0.5, 0.5, 0.5) with radius 0.5.
    Triangulate for safety (it should already be triangular, but just to be sure).
    """
    sph = pv.Sphere(radius=0.5, center=(0.5, 0.5, 0.5), phi_resolution=24, theta_resolution=24)
    return sph.triangulate()

@pytest.fixture
def big_sphere_pv():
    """
    A sphere that fully encompasses the cube: center (0.5,0.5,0.5) and radius=2
    """
    sph = pv.Sphere(radius=2.0, center=(0.5, 0.5, 0.5), phi_resolution=24, theta_resolution=24)
    return sph.triangulate()

@pytest.fixture
def shifted_cube_pv():
    """
    A cube that only partially intersects with the standard unit cube
    by shifting it by 0.5 in x, y, z.
    """
    box = pv.Box(bounds=(0.5, 1.5, 0.5, 1.5, 0.5, 1.5))
    return box.triangulate()

###############################################################################
# Tests for conversion functions
###############################################################################

def test_convert_to_trimesh_triangle():
    """
    Test converting a minimal triangular PyVista mesh to a Trimesh.
    """
    points = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0]
    ])
    faces = np.hstack([[3], [0, 1, 2]])  # single triangle
    pv_mesh = pv.PolyData(points, faces, force_float=False).triangulate()
    tm_mesh = convert_to_trimesh(pv_mesh)
    assert tm_mesh.faces.shape[0] == 1, "Should have exactly 1 face (triangle)."


def test_convert_to_trimesh_quads():
    points = np.array([
        [0,0,0],
        [1,0,0],
        [1,1,0],
        [0,1,0]
    ], dtype=float)
    faces = np.hstack([[4],[0,1,2,3]])  # single quad
    pv_mesh = pv.PolyData(points, faces)
    # Triangulate here:
    pv_mesh = pv_mesh.triangulate()
    # Now this should be 2 triangles => 8 ints in faces
    tm_mesh = convert_to_trimesh(pv_mesh)
    # Should pass, as code can do the standard reshape
    assert tm_mesh.faces.shape[0] == 2, "Quad should become 2 triangles after triangulation."

###############################################################################
# Tests for boolean() function
###############################################################################

def test_boolean_union(cube_pv, sphere_pv):
    """
    Union of a unit cube and a sphere (radius=0.5 inside it).
    The sphere is centered at (0.5,0.5,0.5) so it partially intersects.
    """
    result = boolean(cube_pv, sphere_pv, operation='union')
    assert result.n_cells > 0, "Result should have faces."
    assert result.n_points > 0, "Result should have vertices."
    assert result.is_manifold, "Union result should be manifold if fix_mesh is True."

    # Volume check
    vol_cube = cube_pv.volume
    vol_sphere = sphere_pv.volume
    vol_union = result.volume
    if vol_cube >= vol_sphere:
        test_vol = vol_cube
        test_geo = 'cube'
    else:
        test_vol = vol_sphere
        test_geo = 'sphere'
    # The union should be bigger than or equal to the larger of the two
    assert vol_union >= vol_cube, "Union volume should at least be the {}'s volume.".format(test_geo)
    # The union should be bigger than or equal to the larger of the two
    assert vol_union >= vol_sphere, "Union volume should at least be the {}'s volume.".format(test_geo)
    if not np.isclose(vol_union, test_vol, atol=1e-7):
        # If it's not equal within tolerance, then we expect it to exceed the cube volume
        assert vol_union > test_vol



def test_boolean_intersection(cube_pv, sphere_pv):
    """
    Intersection of a unit cube and a small sphere (radius=0.5).
    The sphere is inside the cube, so intersection ~ the sphere region that
    lies within the cube's boundaries. Should be roughly the sphere volume.
    """
    result = boolean(cube_pv, sphere_pv, operation='intersection')
    assert result.n_cells > 0, "Result should have faces."
    assert result.n_points > 0, "Result should have vertices."
    assert result.is_manifold, "Intersection result should be manifold."

    vol_cube = cube_pv.volume
    vol_sphere = sphere_pv.volume
    vol_intersect = result.volume
    # Because the sphere is entirely inside the cube (since radius=0.5, center at (0.5,0.5,0.5)),
    # the intersection volume should be about the sphere's volume.
    # But to allow for triangulation or mesh approximations, we allow some tolerance
    assert abs(vol_intersect - vol_sphere) / vol_sphere < 0.05, (
        "Intersection volume should be close to the sphere's volume."
    )


def test_boolean_difference(cube_pv, sphere_pv):
    """
    The difference (cube - sphere).
    Because the sphere is fully inside, the difference is the cube shell minus
    that smaller sphere region in the center.
    """
    result = boolean(cube_pv, sphere_pv, operation='difference')
    assert result.n_cells > 0, "Result should have faces."
    assert result.n_points > 0, "Result should have vertices."
    assert result.is_manifold, "Difference result should be manifold."

    vol_cube = cube_pv.volume
    vol_sphere = sphere_pv.volume
    vol_diff = result.volume
    # We removed the inside part from the cube
    # So the difference volume should be about (cube - sphere).
    expected = vol_cube - vol_sphere
    assert abs(vol_diff - expected) / vol_cube < 0.05, (
        "Difference volume should be close to cube - sphere volume."
    )


def test_boolean_difference_outside(cube_pv, big_sphere_pv):
    """
    If we do (cube - big_sphere), and the sphere is much bigger than the cube,
    the difference is effectively empty (the sphere covers the entire cube).
    """
    result = boolean(cube_pv, big_sphere_pv, operation='difference')
    # Possibly the result might be entirely empty. Some boolean engines produce an empty mesh.
    # pyvista might handle an "empty mesh" in different ways (0 faces, 0 points).
    # Let’s see if the code returns a valid but empty PolyData or raises an error.
    assert result.n_cells == 0 or result.n_points == 0, (
        "Expected empty or near-empty result if the sphere fully envelops the cube."
    )
    # If the engine tries to fix the mesh, it might produce a degenerate polydata with no volume
    assert result.volume == 0.0, "Result volume should be zero (fully subtracted)."


def test_boolean_shared_edge(cube_pv, shifted_cube_pv):
    """
    The union of two cubes that share only a partial overlap (shifted by 0.5).
    We can check volumes, bounding box, and face counts more explicitly.
    """
    # They intersect from [0.5..1.0] in each axis, so the overlap region is a sub-cube of size 0.5^3=0.125
    # So total union volume = 1.0 (cube) + 1.0 (shifted cube) - 0.125 (overlap) = 1.875
    result_union = boolean(cube_pv, shifted_cube_pv, operation='union')
    vol_union = result_union.volume
    assert abs(vol_union - 1.875) < 0.05, (
        f"Union volume ~1.875, got {vol_union}"
    )

    # Intersection volume should be ~0.125
    result_intersect = boolean(cube_pv, shifted_cube_pv, operation='intersection')
    vol_intersect = result_intersect.volume
    assert abs(vol_intersect - 0.125) < 0.01, (
        f"Intersection volume ~0.125, got {vol_intersect}"
    )

    # difference (cube - shifted_cube) => volume should be 1.0 - 0.125 = 0.875
    result_diff = boolean(cube_pv, shifted_cube_pv, operation='difference')
    vol_diff = result_diff.volume
    assert abs(vol_diff - 0.875) < 0.05, (
        f"Difference volume ~0.875, got {vol_diff}"
    )


def test_boolean_touching_corner():
    """
    Create two cubes that only share one corner (1,1,1)
    and test intersection is basically that point (which is degenerate).
    """
    cube1 = pv.Box(bounds=(0,1,0,1,0,1)).triangulate()
    # Shift so the second cube goes from (1,2,1,2,1,2),
    # so the only "overlap" is that corner at (1,1,1).
    cube2 = pv.Box(bounds=(1,2,1,2,1,2)).triangulate()

    inter = boolean(cube1, cube2, operation='intersection')
    # Possibly the intersection is an empty/degenerate mesh or a single vertex.
    # Some boolean engines might treat a corner contact as no intersection volume.
    # Let's see if it's empty.
    assert inter.n_cells == 0 or inter.n_points <= 1, (
        "Touching corners produce no real intersection volume."
    )
    # Union should have volume=2
    union = boolean(cube1, cube2, operation='union')
    vol_union = union.volume
    assert abs(vol_union - 2.0) < 0.05, "Two unit cubes union => volume ~2.0 if they only share a corner."


def test_boolean_invalid_operation(cube_pv, sphere_pv):
    """
    Test that an invalid operation raises a ValueError.
    """
    with pytest.raises(ValueError):
        boolean(cube_pv, sphere_pv, operation='bogus_op')



# File: tests/test_minimum_segment_distance.py

import pytest
import numpy as np
from math import sqrt

from svv.utils.spatial.c_distance import minimum_segment_distance

def test_both_degenerate_same_point():
    """
    Both segments are degenerate at the exact same point.
    Distance should be zero.
    """
    data0 = np.array([
        [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
    ], dtype=np.float64)
    data1 = np.array([
        [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(0.0, abs=1e-12)


def test_one_degenerate_one_nondegenerate():
    """
    One segment is a point, the other is a real segment.
    Scenario: point (1,1,1), segment from (2,2,2) to (3,2,2).
    The distance is sqrt((2-1)^2 + (2-1)^2 + (2-1)^2) = sqrt(3)
    if the closest approach is at the segment start.
    """
    data0 = np.array([
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],  # degenerate
    ], dtype=np.float64)
    data1 = np.array([
        [2.0, 2.0, 2.0, 3.0, 2.0, 2.0],  # real segment
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    expected = sqrt(3)
    assert dist[0, 0] == pytest.approx(expected, abs=1e-12)


def test_corner_touch_1d():
    """
    Two collinear segments along x-axis:
    Segment A: [0,0,0 -> 1,0,0]
    Segment B: [1,0,0 -> 2,0,0]

    They share a corner at x=1. The distance is 0 at that shared endpoint.
    """
    data0 = np.array([
        [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],  # A
    ], dtype=np.float64)
    data1 = np.array([
        [1.0, 0.0, 0.0, 2.0, 0.0, 0.0],  # B
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(0.0, abs=1e-12)


def test_partial_overlap_1d():
    """
    Two collinear segments along x-axis that partially overlap:
    A: [0,0,0 -> 2,0,0]
    B: [1,0,0 -> 3,0,0]

    Overlap region is [1,2] along x.
    The distance is 0, because within that overlap region, points coincide.
    """
    data0 = np.array([
        [0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
    ], dtype=np.float64)
    data1 = np.array([
        [1.0, 0.0, 0.0, 3.0, 0.0, 0.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(0.0, abs=1e-12)


def test_parallel_same_plane_offset():
    """
    Parallel segments in the same plane, offset by 5 units along y-axis:
    A: [0,0,0 -> 1,0,0]
    B: [0,5,0 -> 1,5,0]

    Distance is always 5 (the offset).
    """
    data0 = np.array([
        [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
    ], dtype=np.float64)
    data1 = np.array([
        [0.0, 5.0, 0.0, 1.0, 5.0, 0.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(5.0, abs=1e-12)


def test_parallel_different_planes():
    """
    Parallel segments in distinct planes:
    Segment A in plane z=0, Segment B in plane z=2
    Both along x-axis, from x=0..2.

    So the distance is the constant difference in z: 2
    """
    data0 = np.array([
        [0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
    ], dtype=np.float64)
    data1 = np.array([
        [0.0, 0.0, 2.0, 2.0, 0.0, 2.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(2.0, abs=1e-12)


def test_reversed_segments():
    """
    Segments reversed but should yield the same distance as normal orientation.
    A: from (1,1,1) to (0,0,0)
    B: from (2,2,2) to (3,3,3)

    Both diagonals in 3D space. The minimal distance is from (1,1,1) to (2,2,2),
    which is sqrt(3). The orientation reversal should not matter.
    """
    data0 = np.array([
        [1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
    ], dtype=np.float64)
    data1 = np.array([
        [3.0, 3.0, 3.0, 2.0, 2.0, 2.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    expected = sqrt((2 - 1)**2 + (2 - 1)**2 + (2 - 1)**2)  # sqrt(3)
    assert dist[0, 0] == pytest.approx(expected, abs=1e-12)


def test_line_crossing_at_endpoint():
    """
    A crosses B exactly at B's endpoint:
    A: from (0,0,0) to (2,2,0)
    B: from (2,2,0) to (2,2,2)

    The segments meet at (2,2,0). Minimum distance is 0.
    """
    data0 = np.array([
        [0.0, 0.0, 0.0, 2.0, 2.0, 0.0],
    ], dtype=np.float64)
    data1 = np.array([
        [2.0, 2.0, 0.0, 2.0, 2.0, 2.0],
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (1, 1)
    assert dist[0, 0] == pytest.approx(0.0, abs=1e-12)


def test_multiple_pairs():
    """
    A test with multiple segments on each side, to verify shape correctness and
    distances between each pair.

    We'll carefully choose them so that we can reason about expected distances:
    - data0 has 2 segments: S0, S1
    - data1 has 2 segments: T0, T1

    We'll compute the 2x2 matrix of distances and check each entry.
    """
    data0 = np.array([
        [0,0,0, 0,0,1],    # S0: vertical from (0,0,0) to (0,0,1)
        [0,0,2, 1,0,2],    # S1: horizontal from (0,0,2) to (1,0,2)
    ], dtype=np.float64)

    data1 = np.array([
        [0,1,0, 0,2,0],    # T0: vertical from (0,1,0) to (0,2,0)
        [1,0,2, 2,0,2],    # T1: horizontal from (1,0,2) to (2,0,2)
    ], dtype=np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (2, 2)

    # dist[0, 0] => S0 vs T0
    #   S0: z from 0 to 1, y=0, x=0
    #   T0: y from 1 to 2, x=0, z=0
    # The closest approach is at S0 endpoint (0,0,0) to T0 endpoint (0,1,0).
    # Distance = 1
    assert dist[0, 0] == pytest.approx(1.0, abs=1e-12)

    # dist[0, 1] => S0 vs T1
    #   S0: (0,0,0)->(0,0,1); T1: (1,0,2)->(2,0,2)
    #   S0 is along z in the plane x=0, y=0
    #   T1 is along x in plane y=0, z=2
    # The closest approach might be from (0,0,1) to (1,0,2).
    # That distance is sqrt((1-0)^2 + (0-0)^2 + (2-1)^2) = sqrt(1 + 1) = sqrt(2)
    # Let's confirm that’s indeed the min.
    assert dist[0, 1] == pytest.approx(sqrt(2), abs=1e-12)

    # dist[1, 0] => S1 vs T0
    #   S1: (0,0,2)->(1,0,2)
    #   T0: (0,1,0)->(0,2,0)
    #   S1 is along x in plane y=0, z=2
    #   T0 is along y in plane x=0, z=0
    # The segments are quite separated in z. The closest approach might be
    # from S1 endpoint (0,0,2) to T0 endpoint (0,1,0)?
    # That’s distance sqrt((0-0)^2 + (1-0)^2 + (0-2)^2) = sqrt(0 +1 +4) = sqrt(5).
    # It's indeed minimal because the other combos are bigger.
    assert dist[1, 0] == pytest.approx(sqrt(5), abs=1e-12)

    # dist[1, 1] => S1 vs T1
    #   S1: (0,0,2)->(1,0,2), T1: (1,0,2)->(2,0,2)
    # They share a corner at (1,0,2). Distance = 0.
    assert dist[1, 1] == pytest.approx(0.0, abs=1e-12)


def test_many_random_small():
    """
    Random test with a small number of segments to ensure
    the function doesn't crash or produce nonsensical results (NaN or negative).

    This won't be an exact correctness check, but we at least confirm that
    distances are non-negative and we can check for outliers.
    """
    rng = np.random.default_rng(42)
    data0 = rng.random((5, 6))
    data1 = rng.random((6, 6))

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (5, 6)
    assert np.all(dist >= 0.0)
    # We don't know exact distances, but let's ensure no NaN or inf.
    assert np.all(np.isfinite(dist))


@pytest.mark.parametrize(
    "n0, n1",
    [
        (10, 10),
        (100, 100),
        (500, 500),
    ]
)
def test_many_random_performance(n0, n1):
    """
    Larger random test to check performance and confirm no crashes/NaNs.
    """
    rng = np.random.default_rng(123)
    data0 = rng.random((n0, 6)).astype(np.float64)
    data1 = rng.random((n1, 6)).astype(np.float64)

    dist = minimum_segment_distance(data0, data1)
    assert dist.shape == (n0, n1)
    assert np.all(dist >= 0.0)
    assert np.all(np.isfinite(dist)), "Found NaN/inf values in distance result."


import numpy as np
import pyvista as pv
from pyvista import examples
from scipy.interpolate import interp1d
from scipy.spatial.distance import pdist
from scipy.optimize import minimize

from svv.utils.spatial.c_distance import minimum_segment_distance, minimum_self_segment_distance
from svv.forest.connect.curve import Curve

def random_point_in_mesh(mesh: pv.PolyData, max_tries=1000):
    """
    Returns a single random point strictly inside a watertight mesh
    using rejection sampling within its bounding box.
    Raises a ValueError if it fails to find a point after `max_tries`.
    """
    # Get bounding box of the mesh
    xmin, xmax, ymin, ymax, zmin, zmax = mesh.bounds

    # Sample a random point in the bounding box
    x = np.random.uniform(xmin, xmax, max_tries)
    y = np.random.uniform(ymin, ymax, max_tries)
    z = np.random.uniform(zmin, zmax, max_tries)
    candidate = np.array([x, y, z]).T  # shape (1, 3)

    # Convert to PyVista object
    candidate_points = pv.PolyData(candidate)

    # Check if point is inside using `select_enclosed_points`
    enclosed = candidate_points.select_enclosed_points(mesh)
    if np.any(enclosed["SelectedPoints"] == 1):  # 1 means inside
        return candidate_points.extract_points(enclosed['SelectedPoints'].view(bool))  # return as a 1D array [x, y, z]
    else:
        raise ValueError("Could not find a random interior point within max_tries.")


def random_line_segment_in_mesh(mesh: pv.PolyData,
                                resolution=10,
                                max_tries=1000):
    """
    Returns two endpoints (p1, p2) that lie fully inside `mesh` (watertight).
    The entire line from p1 -> p2 is checked to ensure all sampled points
    are inside. Uses your updated random_point_in_mesh() function to pick
    endpoints from multiple inside points at once.

    Parameters
    ----------
    mesh : pv.PolyData
        A watertight PyVista mesh.
    resolution : int
        Number of points along the line segment for the inside check.
    max_tries : int
        Maximum number of attempts to find a fully enclosed line segment.

    Returns
    -------
    (p1, p2) : tuple of np.ndarray
        Each is a 3D coordinate array representing one endpoint of the segment.

    Raises
    ------
    ValueError
        If no fully enclosed line segment could be found in `max_tries` attempts.
    """
    for attempt in range(max_tries):
        # -----------------------------------------------------------
        # 1) Get a set of random points inside the mesh for endpoint 1
        # -----------------------------------------------------------
        try:
            inside_points_1 = random_point_in_mesh(mesh, max_tries=max_tries)
            # 'inside_points_1' is a pv.PolyData with one or more points
            n_inside_1 = inside_points_1.n_points
            # Randomly choose one point among them
            idx1 = np.random.randint(n_inside_1)
            p1 = inside_points_1.points[idx1]
        except ValueError:
            # Could not find a single inside point within max_tries
            continue

        # -----------------------------------------------------------
        # 2) Get a set of random points inside the mesh for endpoint 2
        # -----------------------------------------------------------
        try:
            inside_points_2 = random_point_in_mesh(mesh, max_tries=max_tries)
            n_inside_2 = inside_points_2.n_points
            idx2 = np.random.randint(n_inside_2)
            p2 = inside_points_2.points[idx2]
        except ValueError:
            # Could not find a single inside point within max_tries
            continue

        # If p1 == p2 (or extremely close), skip
        if np.allclose(p1, p2, atol=1e-12):
            continue

        # -----------------------------------------------------------
        # 3) Discretize the line between p1 and p2
        # -----------------------------------------------------------
        t_values = np.linspace(0, 1, resolution)
        line_points_arr = (1.0 - t_values)[:, None] * p1 + t_values[:, None] * p2
        line_points = pv.PolyData(line_points_arr)

        # -----------------------------------------------------------
        # 4) Check if *all* discretized points are inside the mesh
        # -----------------------------------------------------------
        enclosed = line_points.select_enclosed_points(mesh)
        inside_mask = enclosed["SelectedPoints"]  # array of 0 or 1

        if np.all(inside_mask == 1):
            # Found a line segment fully inside
            return p1, p2

    # If we exhaust all tries without finding a fully inside segment, raise
    raise ValueError("Failed to find a line segment fully inside the mesh within max_tries.")
def random_line_segments_in_mesh(mesh: pv.PolyData,
                                 n_segments=5,
                                 resolution=10,
                                 max_tries=2000):
    """
    Generates and returns `n_segments` random line segments
    lying fully inside a watertight mesh.

    Returns
    -------
    segments : list of tuples
        Each tuple is (p1, p2), where both p1 and p2 are
        3-element np.array of coordinates.
    """
    segments = []
    for _ in range(n_segments):
        p1, p2 = random_line_segment_in_mesh(mesh, resolution, max_tries)
        segments.append((p1, p2))
    return segments

class BaseConnection:
    """
    A base connection class that computes an objective function and
    various geometric constraints for an optimized curve connecting
    four points (A->B and C->D) subject to optional clamp constraints.
    """
    def __init__(
        self,
        point_00,
        point_01,
        point_10,
        point_11,
        radius_0,
        radius_1,
        domain=None,
        ctrlpt_function=None,
        clamp_first=True,
        clamp_second=True,
        point_0=None,
        point_1=None,
        min_distance=1e-4,
        clearance=0.0,
        curve_type='Bezier'
    ):
        # Radii
        self.radius_0 = radius_0
        self.radius_1 = radius_1

        # Assign A, B, C, D
        # If point_0 is given, use that for A & B; otherwise, use point_00, point_01.
        if point_0 is not None:
            self.A = point_0
            self.B = point_1
        else:
            self.A = point_00
            self.B = point_01

        # If point_1 is given, use that for C & D; otherwise, use point_10, point_11.
        if point_1 is not None:
            self.C = point_1
            self.D = point_1
        else:
            self.C = point_10
            self.D = point_11

        # Store line segments, domain, etc.
        self.other_line_segments = []
        self.min_distance = min_distance
        self.ctrlpt_function = ctrlpt_function
        self.clamp_first = clamp_first
        self.clamp_second = clamp_second
        self.domain = domain
        self.clearance = clearance
        self.curve_type = curve_type

        # Domain bounding box
        if domain is not None:
            mins = np.min(domain.points, axis=0) * 2
            maxs = np.max(domain.points, axis=0) * 2
            self.x_min, self.x_max = mins[0], maxs[0]
            self.y_min, self.y_max = mins[1], maxs[1]
            self.z_min, self.z_max = mins[2], maxs[2]
        else:
            self.x_min, self.x_max = 0, 1
            self.y_min, self.y_max = 0, 1
            self.z_min, self.z_max = 0, 1

        # Keep track of control points and curves for debugging/analysis
        self.history_ctrlpts = []
        self.history_curves = []

    def _build_control_points(self, ctrlpts_flat, mid, n_pts):
        """
        Construct the full array of control points from the flattened
        optimization variables, taking into account any clamping.
        """
        # Precompute direction unit vectors if clamps are active
        if self.clamp_first:
            BA_dir = (self.B - self.A)
            BA_dir /= np.linalg.norm(BA_dir)
        if self.clamp_second:
            DC_dir = (self.D - self.C)
            DC_dir /= np.linalg.norm(DC_dir)

        # P0 is a mix between A and B; similarly for P3 (mix between C and D)
        P0 = self.A * (1 - ctrlpts_flat[0]) + self.B * ctrlpts_flat[0]
        if self.clamp_first:
            P1 = P0 + ctrlpts_flat[1] * BA_dir

        P3 = self.C * (1 - ctrlpts_flat[-1]) + self.D * ctrlpts_flat[-1]
        if self.clamp_second:
            P2 = P3 + ctrlpts_flat[-2] * DC_dir

        # Midpoints from the 'mid' function
        mid_points = mid(np.linspace(0.25, 0.75, n_pts))

        # Decide how many in the middle (depending on clamps)
        if self.clamp_first and self.clamp_second:
            ctrlpts_middle = mid_points + ctrlpts_flat[2:-2].reshape(-1, 3)
            all_points = np.vstack([P0, P1, ctrlpts_middle, P2, P3])

        elif self.clamp_first:
            ctrlpts_middle = mid_points + ctrlpts_flat[2:-1].reshape(-1, 3)
            all_points = np.vstack([P0, P1, ctrlpts_middle, P3])

        elif self.clamp_second:
            ctrlpts_middle = mid_points + ctrlpts_flat[1:-2].reshape(-1, 3)
            all_points = np.vstack([P0, ctrlpts_middle, P2, P3])

        else:
            ctrlpts_middle = mid_points + ctrlpts_flat[1:-1].reshape(-1, 3)
            all_points = np.vstack([P0, ctrlpts_middle, P3])

        return all_points

    def create_objective(self, t_num, mid, n_pts):
        """
        Returns the objective function (callable) that takes in a
        flattened array of control points and returns a scalar.
        """
        def objective(control_points_flat):
            # Build the control points and the corresponding curve
            control_points = self._build_control_points(control_points_flat, mid, n_pts)
            curve = Curve(control_points, curve_type=self.curve_type)

            # For debugging/analysis
            self.history_ctrlpts.append(control_points)
            self.history_curves.append(curve)

            # Example objective: difference between curve length and some nominal length
            # length = np.linalg.norm(self.B - self.A) + np.linalg.norm(self.D - self.C)
            # or you can do something else:
            length = np.linalg.norm(self.D - self.B)
            return max(0.0, curve.arc_length(num_points=t_num) - length)

        return objective

    def create_constraints(self,radius1,radius2,other_line_segments,t_num,mid,n_pts):
        """
        Returns a list of constraint dictionaries to be used in a
        typical optimizer (e.g., scipy.optimize).
        """

        def curvature_constraint(ctrlpts_flat):
            control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
            curve = Curve(control_points, curve_type=self.curve_type)

            # Evaluate radius of curvature
            t_values = np.linspace(0, 1, 100)
            roc_values = curve.roc(t_values)

            # Enforce min radius of curvature
            min_radius_needed = 2 * max(radius1, radius2)
            return np.min(roc_values[1:-1]) - min_radius_needed

        def non_coincidence_constraint(ctrlpts_flat):
            """
            Ensures that no two control points coincide
            (enforce a small min distance between them).
            """
            control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
            distances = pdist(control_points)
            return np.min(distances) - 1e-4  # or self.min_distance

        def curve_min_distance_constraint(ctrlpts_flat):
            """
            Ensures that the constructed curve is not too close
            to other line segments in the scene.
            """
            control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
            if len(other_line_segments) == 0:
                return 1.0  # If no other segments, no penalty

            curve = Curve(control_points, curve_type=self.curve_type)
            t_values = np.linspace(0, 1, t_num)
            curve_points = curve.evaluate(t_values)

            # Build array of segments from the discretized curve
            segments = np.zeros((curve_points.shape[0] - 1, 6))
            segments[:, :3] = curve_points[:-1]
            segments[:, 3:] = curve_points[1:]

            # Evaluate the distance
            # The 'minimum_segment_distance' presumably returns NxM array
            # of distances between two sets of line segments
            if len(self.other_line_segments) > 0:
                dist_main = np.min(
                    minimum_segment_distance(self.other_line_segments[:, :6], segments)
                    - self.other_line_segments[:, 6].reshape(-1, 1) - max(self.radius_0, self.radius_1)
                ) - self.min_distance
                #dist_main_check = np.min(
                #    cylinders_collide_any_naive(self.other_line_segments[:, :6], segments)
                #    - self.other_line_segments[:, 6].reshape(-1, 1) - max(self.radius_0, self.radius_1)
                #) - self.min_distance
                #assert dist_main == dist_main_check, "{} != {} INCORRECT CHECK".format(dist_main,dist_main_check)
                return dist_main
            else:
                return 1.0

        def boundary_constraint(ctrlpts_flat):
            """
            Ensures that the entire curve lies within the domain
            (with an optional clearance).
            """
            if self.domain is None:
                return 1.0  # No domain => no constraint

            control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
            curve = Curve(control_points, curve_type=self.curve_type)
            t_values = np.linspace(0, 1, 100)
            curve_points = curve.evaluate(t_values)

            # ------------------------------------------------------
            # 1) If domain is a callable, use the original logic
            # ------------------------------------------------------
            if callable(self.domain):
                # domain(curve_points) presumably returns distances to the boundary
                # (negative => out-of-domain, positive => inside)
                values = self.domain(curve_points)
                # We want the entire curve to be inside =>
                #   the maximum distance should remain positive
                # The additional clearance is subtracted to enforce a "buffer zone".
                return -(np.max(values) + self.clearance)

            # ------------------------------------------------------
            # 2) If domain is PyVista PolyData, do an inside test
            # ------------------------------------------------------
            elif isinstance(self.domain, pv.PolyData):
                # We assume 'self.domain' is a closed (watertight) mesh.
                # We'll create a temporary PyVista mesh from the curve points
                temp_points = pv.PolyData(curve_points)

                # Use the select_enclosed_points filter:
                #   `enclosed_result` is typically an UnstructuredGrid with a point-data
                #   array named 'SelectedPoints' (1 = inside, 0 = outside).
                enclosed_result = temp_points.select_enclosed_points(self.domain, tolerance=0.0)
                inside_mask = enclosed_result['SelectedPoints']  # numpy array of 0 or 1

                if not np.all(inside_mask):
                    # Some points of the curve lie outside the domain surface
                    return -1.0
                else:
                    # Entire curve is inside
                    # If you also want a clearance, consider measuring the distance
                    # from each point to the surface (using e.g. 'find_closest_point')
                    # and ensuring it exceeds self.clearance. That logic would replace
                    # or supplement the simple "inside test" here.
                    return 1.0

            # ------------------------------------------------------
            # 3) Fallback if domain is of another type (optional)
            # ------------------------------------------------------
            else:
                # You could raise an error or just return a non-violating value
                # depending on your application's needs.
                print("Warning: Domain type not recognized. No boundary constraint applied.")
                return 1.0

        # You could also add a constraint on the raw control points
        # bounding box if desired:
        #
        # def ctrlpts_boundary_constraint(ctrlpts_flat):
        #     control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
        #     ctrl_min = np.min(control_points, axis=0)
        #     ctrl_max = np.max(control_points, axis=0)
        #     # Must be within [x_min, x_max], [y_min, y_max], [z_min, z_max]
        #     # Return min(...) so that if anything is out of bounds, constraint < 0
        #     return min(
        #         *(ctrl_min - [self.x_min, self.y_min, self.z_min]),
        #         *([self.x_max, self.y_max, self.z_max] - ctrl_max)
        #     )
        def self_collision(ctrlpts_flat):
            control_points = self._build_control_points(ctrlpts_flat, mid, n_pts)
            curve = Curve(control_points, curve_type=self.curve_type)
            t_values = np.linspace(0, 1, t_num)
            curve_points = curve.evaluate(t_values)
            segments = np.zeros((t_num - 1, 7))
            segments[:, :3] = curve_points[:-1]
            segments[:, 3:6] = curve_points[1:]
            segments[:, 6] = max(self.radius_0, self.radius_1)
            dist_main = np.min(minimum_self_segment_distance(segments[:, :6]) - max(self.radius_0, self.radius_1)) - self.min_distance
            return dist_main

        return [
            {'type': 'ineq', 'fun': curvature_constraint},
            {'type': 'ineq', 'fun': curve_min_distance_constraint},
            {'type': 'ineq', 'fun': non_coincidence_constraint},
            {'type': 'ineq', 'fun': boundary_constraint},
            #{'type': 'ineq', 'fun': self_collision}
            # Example if you also want the control-points bounding box check:
            # {'type': 'ineq', 'fun': ctrlpts_boundary_constraint},
        ]

    def solve(self, n_mid_pts, t_num=20):
        """
        Solves for an optimized curve connecting A->B and C->D using
        an objective function and constraints. Clamping can be applied
        to the start/end tangents if desired (clamp_first/clamp_second).

        Parameters
        ----------
        n_mid_pts : int
            Number of 'middle' control points used in the curve construction.
            This was originally args[0].
        t_num : int
            Number of points used for discretizing/evaluating the curve
            for the objective and constraints.

        Returns
        -------
        result : OptimizeResult
            The result returned by scipy.optimize.minimize (includes x, success, etc.).
        curve : Curve
            The final Curve object constructed from the optimized control points.
        """
        # 1) Build an interpolation ('mid') function from B->D
        #    if no custom ctrlpt_function is provided.
        pts_bd = np.vstack((self.B, self.D))
        xpts = interp1d([0, 1], pts_bd[:, 0], kind='linear')
        ypts = interp1d([0, 1], pts_bd[:, 1], kind='linear')
        zpts = interp1d([0, 1], pts_bd[:, 2], kind='linear')

        if self.ctrlpt_function is None or isinstance(self.ctrlpt_function, list):
            # By default, just linearly interpolate between B and D
            def mid(t):
                return np.array([xpts(t), ypts(t), zpts(t)]).T
        else:
            # Use a custom function (presumably expects t in [0,1])
            mid = self.ctrlpt_function

        # 2) Build the objective and constraints
        objective = self.create_objective(t_num, mid, n_mid_pts)
        constraints = self.create_constraints(
            self.radius_0,
            self.radius_1,
            self.other_line_segments,
            t_num,
            mid,
            n_mid_pts
        )

        # 3) Create an initial guess for the optimization variables
        #    - Format: [ clamp_first_alpha, clamp_first_dist, (3*n_mid_pts) mid devs..., clamp_second_dist, clamp_second_alpha ]
        #    - The shape is n_mid_pts * 3 (x,y,z per middle point) + 4
        #      (two scalars for clamp_first, two scalars for clamp_second).
        initial_control_points = np.zeros(n_mid_pts * 3 + 4)
        # Some example initial values for the clamps:
        initial_control_points[0] = 0.85  # clamp_first_alpha
        initial_control_points[1] = 1.0  # clamp_first_dist
        initial_control_points[-2] = 1.0  # clamp_second_dist
        initial_control_points[-1] = 0.85  # clamp_second_alpha

        # 4) Construct bounds for each parameter
        length = np.linalg.norm(self.B - self.D)
        lb = [-length] * initial_control_points.size
        ub = [length] * initial_control_points.size

        # Example clamp bounds
        lb[0], ub[0] = 0.25, 1.0  # clamp_first_alpha ∈ [0.5, 1]
        lb[1], ub[1] = 0.1 * length, 10 * length  # clamp_first_dist  ∈ [0.1*L, 10*L]
        lb[-1], ub[-1] = 0.25, 1.0  # clamp_second_alpha ∈ [0.5, 1]
        lb[-2], ub[-2] = 0.1 * length, 10 * length  # clamp_second_dist ∈ [0.1*L, 10*L]

        # 5) Remove clamp-related parameters if not used
        #    (i.e., if clamp_first=False and/or clamp_second=False).
        if self.clamp_first and self.clamp_second:
            pass  # keep all four clamp variables
        elif self.clamp_first:
            # Remove the second-to-last element (clamp_second_dist)
            # because we are not using the second clamp's tangent distance
            initial_control_points = np.delete(initial_control_points, -2)
            del lb[-2], ub[-2]
        elif self.clamp_second:
            # Remove the second element (clamp_first_dist)
            # because we are not using the first clamp's tangent distance
            initial_control_points = np.delete(initial_control_points, 1)
            del lb[1], ub[1]
        else:
            # Remove both clamp distances (second-to-last and the second element)
            initial_control_points = np.delete(initial_control_points, [-2, 1])
            del lb[-2], ub[-2]
            del lb[1], ub[1]

        bounds = list(zip(lb, ub))

        # 6) Optimize
        result = minimize(objective,initial_control_points,method='SLSQP',constraints=constraints,bounds=bounds)

        # 7) Reconstruct the final control points from the optimized result
        #    - We'll replicate the logic used in _build_control_points or
        #      expand it here in-line.

        opt_x = result.x  # For clarity

        if self.clamp_first:
            BA = (self.B - self.A)
            BA /= np.linalg.norm(BA)
        if self.clamp_second:
            DC = (self.D - self.C)
            DC /= np.linalg.norm(DC)

        # P0: blend of A and B by alpha
        P0 = self.A * (1 - opt_x[0]) + self.B * opt_x[0]
        # If clamp_first, we have a distance param for the tangent
        if self.clamp_first:
            P1 = P0 + opt_x[1] * BA

        # P3: blend of C and D by alpha
        P3 = self.C * (1 - opt_x[-1]) + self.D * opt_x[-1]
        # If clamp_second, we have a distance param for that tangent
        if self.clamp_second:
            P2 = P3 + opt_x[-2] * DC

        # Middle control points
        mid_points = mid(np.linspace(0.25, 0.75, n_mid_pts))

        # Indices for the middle portion depend on clamps being present
        if self.clamp_first and self.clamp_second:
            middle_flat = opt_x[2:-2]
        elif self.clamp_first:
            middle_flat = opt_x[2:-1]
        elif self.clamp_second:
            middle_flat = opt_x[1:-2]
        else:
            middle_flat = opt_x[1:-1]

        middle_deviations = middle_flat.reshape(-1, 3)
        control_points = mid_points + middle_deviations

        # Stack everything together
        if self.clamp_first and self.clamp_second:
            control_points = np.vstack([P0, P1, control_points, P2, P3])
        elif self.clamp_first:
            control_points = np.vstack([P0, P1, control_points, P3])
        elif self.clamp_second:
            control_points = np.vstack([P0, control_points, P2, P3])
        else:
            control_points = np.vstack([P0, control_points, P3])

        # 8) Build final curve and return
        curve = Curve(control_points, curve_type=self.curve_type)
        self.history_curves.append(curve)
        return result, curve, objective, constraints

    def set_collision_vessels(self, collision_vessels):
        self.other_line_segments = collision_vessels

    def set_physical_clearance(self, clearance):
        self.clearance = clearance

    def plot(self, n_points=50, plot_as_tube=True):
        """
        Plots the final optimized connection curve (if it exists) and any
        additional line segments stored in `self.other_line_segments` as cylinders.

        Parameters
        ----------
        n_points : int, optional
            Number of discrete points to use when sampling the final curve
            for visualization.
        plot_as_tube : bool, optional
            If True, the optimized curve is also displayed as a tube with a
            small radius. Otherwise, it's drawn as a simple line.
        """
        # 1) Check if a final solution (curve) is available.
        if not self.history_curves:
            print("No solution found yet. Please run solve() first.")
            return
        # 2) Get the final curve from the history and discretize it
        final_curve = self.history_curves[-1]
        t_vals = np.linspace(0, 1, n_points)
        curve_points = final_curve.evaluate(t_vals)
        # 3) Initialize a PyVista plotter
        plotter = pv.Plotter()
        if isinstance(self.domain, pv.PolyData):
            plotter.add_mesh(self.domain, opacity=0.25)
        # 4) Plot the final curve
        if plot_as_tube:
            # Create a PolyData line, then use .tube(...) to make it a tubular mesh
            vessels = np.zeros((curve_points.shape[0]-1, 7))
            vessels[:, 0:3] = curve_points[:-1, :]
            vessels[:, 3:6] = curve_points[1:, :]
            vessels[:, 6] = max(self.radius_0, self.radius_1)
            meshes = []
            for i in range(vessels.shape[0]):
                center = (vessels[i, 0:3] + vessels[i, 3:6])/2
                direction = (vessels[i, 3:6] - vessels[i, 0:3])
                length = np.linalg.norm(direction)
                direction = direction/length
                cyl = pv.Cylinder(radius = (self.radius_0 + self.radius_1)/2, center = center, direction = direction,
                                  height=length, capping=True)
                meshes.append(cyl)
            tube = pv.merge(meshes)
            plotter.add_mesh(tube, color='blue', label='Optimized Curve')
            #plot first vessel
            center = (vessels[0, 0:3] + self.A)/2
            direction = (vessels[0, 0:3] - self.A)
            length = np.linalg.norm(direction)
            direction = direction/length
            cyl = pv.Cylinder(radius=(self.radius_0 + self.radius_1) / 2, center=center, direction=direction,
                              height=length, capping=True)
            plotter.add_mesh(cyl, color='green', label='Start')
            center = (vessels[-1, 3:6] + self.C)/2
            direction = (self.C - vessels[-1, 3:6])
            length = np.linalg.norm(direction)
            direction = direction/length
            cyl = pv.Cylinder(radius=(self.radius_0 + self.radius_1) / 2, center=center, direction=direction,
                              height=length, capping=True)
            plotter.add_mesh(cyl, color='green', label='End')

        else:
            # Just plot as a simple line
            curve_polydata = pv.lines_from_points(curve_points)
            plotter.add_mesh(curve_polydata, color='blue', line_width=4, label='Optimized Curve')

        # 5) Plot other line segments as cylinders
        #    We assume each segment is [x1, y1, z1, x2, y2, z2, radius].
        #    If the radius is missing, we default to a small number.
        if len(self.other_line_segments) > 0:
            for i, seg in enumerate(self.other_line_segments):
                start = np.array(seg[0:3])
                end = np.array(seg[3:6])

                # If segment has a 7th element, treat that as radius
                if len(seg) >= 7:
                    radius = seg[6]
                else:
                    radius = 0.01  # Fallback radius

                # Compute cylinder parameters:
                direction = end - start
                height = np.linalg.norm(direction)
                # If start and end coincide, skip plotting or treat as a small segment
                if np.isclose(height, 0):
                    continue
                direction /= height  # Normalize for PyVista's cylinder

                center = 0.5 * (start + end)  # Midpoint in world space
                cylinder = pv.Cylinder(
                    center=center,
                    direction=direction,
                    radius=radius,
                    height=height,
                    resolution=24  # number of radial subdivisions (smoothness)
                )

                plotter.add_mesh(cylinder, color='red', label=f'Segment {i + 1}')

        # 6) Optional: add a legend if you want to show labels in the plot
        #    Note that PyVista only displays unique labels; duplicates are merged.
        plotter.add_legend()

        # 7) Show the interactive PyVista window
        plotter.show()

def test(curve_type="Bezier"):
    cow = examples.download_cow()
    segments = np.array(random_line_segments_in_mesh(cow,n_segments=2)).reshape((4,3))
    other_segments = np.array(random_line_segments_in_mesh(cow, n_segments=20)).reshape((20,6))
    vessels = np.zeros((20,7))
    vessels[:, 0:6] = other_segments
    vessels[:, -1] = 0.05
    conn = BaseConnection(segments[0],segments[1],segments[2],segments[3],0.05,0.05,domain=cow,curve_type=curve_type)
    conn.set_collision_vessels(vessels)
    return conn


def cylinders_collide_any(cylinders_A: np.ndarray, cylinders_B: np.ndarray) -> bool:
    """
    Checks if any pair of cylinders (one from `cylinders_A`, one from `cylinders_B`)
    intersects (collides).

    The inputs `cylinders_A` and `cylinders_B` are (N,7) and (M,7) arrays respectively, where:
      - columns 0..5: endpoints of the cylinder's central axis
      - column 6: radius

    Returns:
      True if at least one pair of cylinders is colliding; otherwise False.
    """
    # 1) Extract just the first 6 columns for the line segments
    segments_A = cylinders_A[:, :6]  # shape (N, 6)
    segments_B = cylinders_B[:, :6]  # shape (M, 6)
    # 2) Extract the radii
    radii_A = cylinders_A[:, 6]  # shape (N,)
    radii_B = cylinders_B[:, 6]  # shape (M,)
    # 3) Compute the pairwise distances between line segments (N x M).
    #    You should already have a function like `minimum_segment_distance(segA, segB)`
    #    that returns a (N,M) array of distances.
    dist_array = minimum_segment_distance(segments_A, segments_B)  # shape (N, M)
    # 4) Compute sum of radii for each pair.
    #    We can expand dims to broadcast properly: (N,1) + (1,M) => (N,M)
    sum_of_radii = radii_A[:, None] + radii_B[None, :]
    # 5) Compare distances vs. sum_of_radii. A collision occurs if distance < sum_of_radii.
    #    So if `np.any(dist < sum_of_radii)` is True => at least one collision.
    collisions_mask = dist_array < sum_of_radii
    return np.any(collisions_mask)

def cylinders_collide_any_naive(cylinders_A: np.ndarray, cylinders_B: np.ndarray):
    """
    Checks if any pair of cylinders (one from `cylinders_A`, one from `cylinders_B`)
    collides, by computing the minimal distance between each pair of line segments
    and comparing it to the sum of their radii.

    Parameters
    ----------
    cylinders_A : np.ndarray
        Shape (N,7). Columns [x1, y1, z1, x2, y2, z2, radius].
    cylinders_B : np.ndarray
        Shape (M,7). Columns [x1, y1, z1, x2, y2, z2, radius].

    Returns
    -------
    bool
        True if at least one pair collides; False otherwise.
    """
    # For each cylinder in A, for each cylinder in B:
    # 1) Get the line segments Aseg, Bseg
    # 2) Compute min distance between those line segments
    # 3) Compare to sum_of_radii
    # If distance < sum_of_radii => collision => return True.
    # If we exit loops with no collisions => return False.
    dists = []
    for a in cylinders_A:
        # A's endpoints
        #Ax1, Ay1, Az1, Ax2, Ay2, Az2, Aradius = a
        Ax1, Ay1, Az1, Ax2, Ay2, Az2 = a
        sub_dists = []
        for b in cylinders_B:
            # B's endpoints
            #Bx1, By1, Bz1, Bx2, By2, Bz2, Bradius = b
            Bx1, By1, Bz1, Bx2, By2, Bz2 = b
            # sum of radii
            #sum_radii = Aradius + Bradius
            # If segment distance < sum_radii => collision
            dist = segment_segment_distance_3d(
                Ax1, Ay1, Az1, Ax2, Ay2, Az2,
                Bx1, By1, Bz1, Bx2, By2, Bz2
            )
            #if dist < sum_radii:
            sub_dists.append(dist)
        dists.append(sub_dists)
    return np.array(dists)


def segment_segment_distance_3d(
    Ax1, Ay1, Az1, Ax2, Ay2, Az2,
    Bx1, By1, Bz1, Bx2, By2, Bz2
) -> float:
    """
    Computes the minimal Euclidean distance between two 3D line segments:
      Segment A: (Ax1,Ay1,Az1) -> (Ax2,Ay2,Az2)
      Segment B: (Bx1,By1,Bz1) -> (Bx2,By2,Bz2)

    Returns
    -------
    float
        The minimal distance between the segments in 3D space.
    """
    # Direction vectors
    ABx = Ax2 - Ax1
    ABy = Ay2 - Ay1
    ABz = Az2 - Az1
    CDx = Bx2 - Bx1
    CDy = By2 - By1
    CDz = Bz2 - Bz1
    # Vector between first endpoints
    # CA = A1 - B1
    CAx = Ax1 - Bx1
    CAy = Ay1 - By1
    CAz = Az1 - Bz1
    # Dot products
    AB_AB = ABx*ABx + ABy*ABy + ABz*ABz    # |AB|^2
    CD_CD = CDx*CDx + CDy*CDy + CDz*CDz    # |CD|^2
    AB_CD = ABx*CDx + ABy*CDy + ABz*CDz    # AB . CD
    CA_AB = CAx*ABx + CAy*ABy + CAz*ABz    # CA . AB
    CA_CD = CAx*CDx + CAy*CDy + CAz*CDz    # CA . CD
    EPS = 1e-14
    # Degenerate checks
    if AB_AB < EPS and CD_CD < EPS:
        # Both segments are points
        dx = Ax1 - Bx1
        dy = Ay1 - By1
        dz = Az1 - Bz1
        return np.sqrt(dx*dx + dy*dy + dz*dz)
    elif AB_AB < EPS:
        # Segment A is a point, distance from that point to segment B
        return point_segment_distance_3d(Ax1, Ay1, Az1, Bx1, By1, Bz1, Bx2, By2, Bz2)
    elif CD_CD < EPS:
        # Segment B is a point
        return point_segment_distance_3d(Bx1, By1, Bz1, Ax1, Ay1, Az1, Ax2, Ay2, Az2)
    # If both are non-degenerate
    denominator = AB_AB * CD_CD - AB_CD * AB_CD
    if abs(denominator) < EPS:
        # The segments are parallel (or nearly so).
        # We can use a fallback approach: check each endpoint vs. the other segment
        # and pick the min among the four point-seg distances.
        dA1 = point_segment_distance_3d(Ax1, Ay1, Az1, Bx1, By1, Bz1, Bx2, By2, Bz2)
        dA2 = point_segment_distance_3d(Ax2, Ay2, Az2, Bx1, By1, Bz1, Bx2, By2, Bz2)
        dB1 = point_segment_distance_3d(Bx1, By1, Bz1, Ax1, Ay1, Az1, Ax2, Ay2, Az2)
        dB2 = point_segment_distance_3d(Bx2, By2, Bz2, Ax1, Ay1, Az1, Ax2, Ay2, Az2)
        return min(dA1, dA2, dB1, dB2)
    else:
        # Skew or intersecting lines, use the param formula
        t = (AB_CD*CA_CD - CA_AB*CD_CD) / denominator
        s = (AB_AB*CA_CD - AB_CD*CA_AB) / denominator
        # Clamp t, s within [0..1]
        if t < 0.0:
            t = 0.0
        elif t > 1.0:
            t = 1.0
        if s < 0.0:
            s = 0.0
        elif s > 1.0:
            s = 1.0
        # Closest points
        # P1 = A1 + t * AB
        P1x = Ax1 + t*ABx
        P1y = Ay1 + t*ABy
        P1z = Az1 + t*ABz
        # P2 = B1 + s * CD
        P2x = Bx1 + s*CDx
        P2y = By1 + s*CDy
        P2z = Bz1 + s*CDz
        dx = P1x - P2x
        dy = P1y - P2y
        dz = P1z - P2z
        return np.sqrt(dx*dx + dy*dy + dz*dz)


def point_segment_distance_3d(
    Px, Py, Pz,
    Sx1, Sy1, Sz1,
    Sx2, Sy2, Sz2
) -> float:
    """
    Distance from point (Px,Py,Pz) to the segment [Sx1,Sy1,Sz1 -> Sx2,Sy2,Sz2].
    """
    vx = Sx2 - Sx1
    vy = Sy2 - Sy1
    vz = Sz2 - Sz1
    wx = Px - Sx1
    wy = Py - Sy1
    wz = Pz - Sz1
    seg_len_sq = vx*vx + vy*vy + vz*vz
    if seg_len_sq < 1e-14:
        # Degenerate => just distance from (Sx1,Sy1,Sz1)
        dx = Px - Sx1
        dy = Py - Sy1
        dz = Pz - Sz1
        return np.sqrt(dx*dx + dy*dy + dz*dz)
    proj = (wx*vx + wy*vy + wz*vz) / seg_len_sq
    if proj < 0.0:
        proj = 0.0
    elif proj > 1.0:
        proj = 1.0
    cx = Sx1 + proj*vx
    cy = Sy1 + proj*vy
    cz = Sz1 + proj*vz
    dx = Px - cx
    dy = Py - cy
    dz = Pz - cz
    return np.sqrt(dx*dx + dy*dy + dz*dz)


def create_cylinder_mesh(p0: np.ndarray, p1: np.ndarray, radius: float) -> pv.PolyData:
    """
    Create a PyVista mesh (PolyData) of a cylinder whose centerline
    goes from p0 to p1, with the specified radius.
    If the segment length is near zero, return a sphere as a fallback.
    """
    # Direction vector
    v = p1 - p0
    length = np.linalg.norm(v)

    # Handle degenerate segment => fallback as a sphere at p0
    if length < 1e-12:
        return pv.Sphere(radius=radius, center=p0)

    direction = v / length
    midpoint = 0.5 * (p0 + p1)

    cylinder = pv.Cylinder(
        center=midpoint,
        direction=direction,
        height=length,
        radius=radius
    )
    return cylinder


def batch_cylinder_intersection(data_0: np.ndarray,
                                data_1: np.ndarray) -> np.ndarray:
    """
    Build cylinder meshes for each row in data_0 and data_1,
    and check pairwise intersections via boolean operations.

    Args:
        data_0, data_1: Arrays of shape (N,7) and (M,7).
                        Each row is (x0, y0, z0, x1, y1, z1, radius).

    Returns:
        A boolean array `intersect[N, M]` where True means cylinders intersect.
    """
    n = data_0.shape[0]
    m = data_1.shape[0]

    # Pre-build the cylinder meshes for data_0
    cyls_0 = []
    for i in range(n):
        x0, y0, z0, x1, y1, z1, r0 = data_0[i]
        cyls_0.append(create_cylinder_mesh(
            np.array([x0, y0, z0]),
            np.array([x1, y1, z1]),
            r0
        ))

    # Pre-build the cylinder meshes for data_1
    cyls_1 = []
    for j in range(m):
        x0, y0, z0, x1, y1, z1, r1 = data_1[j]
        cyls_1.append(create_cylinder_mesh(
            np.array([x0, y0, z0]),
            np.array([x1, y1, z1]),
            r1
        ))

    # Prepare the output matrix
    intersect = np.zeros((n, m), dtype=bool)

    # Pairwise boolean intersection check
    for i in range(n):
        for j in range(m):
            # Attempt a boolean cut or intersection
            # If the result has any cells, they intersect
            res = cyls_0[i].boolean_cut(cyls_1[j])
            intersect[i, j] = (res.n_cells > 0)

    return intersect


# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os

def get_logger_name():
    return 'generate-1d-mesh'

def get_log_file_name():
    return 'generate-1d-mesh.log'

def init_logging(outputDir="./"):
    import logging
    logger = logging.getLogger(get_logger_name())
    logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    formatter = logging.Formatter('[%(name)s] %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    logFile = os.path.join(outputDir, get_log_file_name())
    file_handler = logging.FileHandler(logFile, mode="w")
    logger.addHandler(file_handler)
    file_handler.setFormatter(formatter)

